#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import copy
from unittest import mock

from octavia_lib.api.drivers import data_models
from octavia_lib.api.drivers import driver_lib as o_driver_lib
from octavia_lib.api.drivers import exceptions
from octavia_lib.common import constants
import openstack
from oslo_utils import uuidutils
from ovsdbapp.backend.ovs_idl import idlutils

from ovn_octavia_provider.common import clients
from ovn_octavia_provider.common import constants as ovn_const
from ovn_octavia_provider.common import exceptions as ovn_exc
from ovn_octavia_provider import driver as ovn_driver
from ovn_octavia_provider import helper as ovn_helper
from ovn_octavia_provider.tests.unit import base as ovn_base
from ovn_octavia_provider.tests.unit import fakes


class TestOvnProviderDriver(ovn_base.TestOvnOctaviaBase):

    def setUp(self):
        super().setUp()
        self.driver = ovn_driver.OvnProviderDriver()
        add_req_thread = mock.patch.object(ovn_helper.OvnProviderHelper,
                                           'add_request')
        self.member_line = (
            'member_%s_%s:%s_%s' %
            (self.member_id, self.member_address,
             self.member_port, self.member_subnet_id))
        self.member_line_additional_vips = (
            'member_%s_%s:%s_%s' %
            (self.member_id, self.member_address,
             self.member_port, self.member_subnet_id))
        self.ovn_lb = mock.MagicMock()
        self.ovn_lb.name = 'foo_ovn_lb'
        self.fake_vip = '10.22.33.4'
        self.ovn_lb.external_ids = {
            ovn_const.LB_EXT_IDS_VIP_KEY: self.fake_vip,
            'pool_%s' % self.pool_id: self.member_line,
            'listener_%s' % self.listener_id: '80:pool_%s' % self.pool_id}
        self.ovn_lb_addi_vips = mock.MagicMock()
        self.ovn_lb_addi_vips.name = 'foo_ovn_lb_addi_vips'
        self.ovn_lb_addi_vips.external_ids = {
            ovn_const.LB_EXT_IDS_VIP_KEY: '10.22.33.4',
            ovn_const.LB_EXT_IDS_ADDIT_VIP_KEY: '2001:db8:0:1::203',
            'pool_%s' % self.pool_id: ','.join([
                self.member_line, self.member_line_additional_vips]),
            'listener_%s' % self.listener_id: '80:pool_%s' % self.pool_id}
        self.mock_add_request = add_req_thread.start()
        self.project_id = uuidutils.generate_uuid()

        self.fail_member = data_models.Member(
            address='198.51.100.4',
            admin_state_up=True,
            member_id=self.member_id,
            monitor_address="100.200.200.100",
            monitor_port=66,
            name='Amazin',
            pool_id=self.pool_id,
            protocol_port=99,
            subnet_id=self.member_subnet_id,
            weight=55)
        self.ref_member = data_models.Member(
            address='198.52.100.4',
            admin_state_up=True,
            member_id=self.member_id,
            monitor_address=data_models.Unset,
            monitor_port=data_models.Unset,
            name='Amazing',
            pool_id=self.pool_id,
            protocol_port=99,
            subnet_id=self.member_subnet_id,
            weight=55)
        self.update_member = data_models.Member(
            address='198.53.100.4',
            admin_state_up=False,
            member_id=self.member_id,
            monitor_address=data_models.Unset,
            monitor_port=data_models.Unset,
            name='Amazin',
            pool_id=self.pool_id,
            protocol_port=99,
            subnet_id=self.member_subnet_id,
            weight=55)
        self.ref_update_pool = data_models.Pool(
            admin_state_up=False,
            description='pool',
            name='Peter',
            lb_algorithm=constants.LB_ALGORITHM_SOURCE_IP_PORT,
            loadbalancer_id=self.loadbalancer_id,
            listener_id=self.listener_id,
            members=[self.ref_member],
            pool_id=self.pool_id,
            protocol='TCP',
            session_persistence={'type': 'SOURCE_IP'})
        self.ref_pool = data_models.Pool(
            admin_state_up=True,
            description='pool',
            name='Peter',
            lb_algorithm=constants.LB_ALGORITHM_SOURCE_IP_PORT,
            loadbalancer_id=self.loadbalancer_id,
            listener_id=self.listener_id,
            members=[self.ref_member],
            pool_id=self.pool_id,
            protocol='TCP',
            session_persistence={'type': 'SOURCE_IP'})
        self.ref_http_pool = data_models.Pool(
            admin_state_up=True,
            description='pool',
            lb_algorithm=constants.LB_ALGORITHM_SOURCE_IP_PORT,
            loadbalancer_id=self.loadbalancer_id,
            listener_id=self.listener_id,
            members=[self.ref_member],
            name='Groot',
            pool_id=self.pool_id,
            protocol='HTTP',
            session_persistence={'type': 'fix'})
        self.ref_lc_pool = data_models.Pool(
            admin_state_up=True,
            description='pool',
            lb_algorithm=constants.LB_ALGORITHM_LEAST_CONNECTIONS,
            loadbalancer_id=self.loadbalancer_id,
            listener_id=self.listener_id,
            members=[self.ref_member],
            name='Groot',
            pool_id=self.pool_id,
            protocol='HTTP',
            session_persistence={'type': 'fix'})
        self.ref_listener = data_models.Listener(
            admin_state_up=False,
            connection_limit=5,
            default_pool=self.ref_pool,
            default_pool_id=self.pool_id,
            listener_id=self.listener_id,
            loadbalancer_id=self.loadbalancer_id,
            name='listener',
            protocol='TCP',
            protocol_port=42)
        self.ref_listener_udp = data_models.Listener(
            admin_state_up=False,
            connection_limit=5,
            default_pool=self.ref_pool,
            default_pool_id=self.pool_id,
            listener_id=self.listener_id,
            loadbalancer_id=self.loadbalancer_id,
            name='listener',
            protocol='UDP',
            protocol_port=42)
        self.ref_listener_sctp = data_models.Listener(
            admin_state_up=False,
            connection_limit=5,
            default_pool=self.ref_pool,
            default_pool_id=self.pool_id,
            listener_id=self.listener_id,
            loadbalancer_id=self.loadbalancer_id,
            name='listener',
            protocol='SCTP',
            protocol_port=42)
        self.fail_listener = data_models.Listener(
            admin_state_up=False,
            connection_limit=5,
            default_pool=self.ref_pool,
            default_pool_id=self.pool_id,
            listener_id=self.listener_id,
            loadbalancer_id=self.loadbalancer_id,
            name='listener',
            protocol='http',
            protocol_port=42)
        self.ref_lb_fully_populated = data_models.LoadBalancer(
            admin_state_up=False,
            listeners=[self.ref_listener],
            pools=[self.ref_pool],
            loadbalancer_id=self.loadbalancer_id,
            name='favorite_lb0',
            project_id=self.project_id,
            vip_address=self.vip_address,
            vip_network_id=self.vip_network_id,
            vip_subnet_id=self.vip_subnet_id)
        self.ref_lb0 = data_models.LoadBalancer(
            admin_state_up=False,
            listeners=[self.ref_listener],
            loadbalancer_id=self.loadbalancer_id,
            name='favorite_lb0',
            project_id=self.project_id,
            vip_address=self.vip_address,
            vip_network_id=self.vip_network_id)
        self.ref_lb1 = data_models.LoadBalancer(
            admin_state_up=True,
            listeners=[self.ref_listener],
            loadbalancer_id=self.loadbalancer_id,
            name='favorite_lb1',
            project_id=self.project_id,
            vip_address=self.vip_address,
            vip_network_id=self.vip_network_id)
        self.ref_lb2 = data_models.LoadBalancer(
            admin_state_up=False,
            listeners=[self.ref_listener],
            loadbalancer_id=self.loadbalancer_id,
            name='additional_vips_lb2',
            project_id=self.project_id,
            vip_address=self.vip_address,
            vip_network_id=self.vip_network_id,
            additional_vips=self.additional_vips)
        self.fail_health_monitor = data_models.HealthMonitor(
            admin_state_up=True,
            name='UnHealthy',
            pool_id=self.pool_id,
            healthmonitor_id=self.healthmonitor_id,
            type="not_valid",
            delay=1,
            timeout=2,
            max_retries_down=3,
            max_retries=4)
        self.ref_health_monitor = data_models.HealthMonitor(
            admin_state_up=True,
            name='Healthy',
            pool_id=self.pool_id,
            healthmonitor_id=self.healthmonitor_id,
            type=constants.HEALTH_MONITOR_TCP,
            delay=6,
            timeout=7,
            max_retries_down=5,
            max_retries=3)
        self.ref_update_health_monitor = data_models.HealthMonitor(
            admin_state_up=True,
            name='ReHealthy',
            healthmonitor_id=self.healthmonitor_id,
            delay=16,
            timeout=17,
            max_retries_down=15,
            max_retries=13)
        self.ref_pool_with_hm = data_models.Pool(
            admin_state_up=True,
            description='pool',
            name='Peter',
            lb_algorithm=constants.LB_ALGORITHM_SOURCE_IP_PORT,
            loadbalancer_id=self.loadbalancer_id,
            listener_id=self.listener_id,
            healthmonitor=self.ref_health_monitor,
            members=[self.ref_member],
            pool_id=self.pool_id,
            protocol='TCP',
            session_persistence={'type': 'SOURCE_IP'})
        self.ref_lb_fully_sync_populated = data_models.LoadBalancer(
            admin_state_up=False,
            listeners=[self.ref_listener],
            pools=[self.ref_pool_with_hm],
            loadbalancer_id=self.loadbalancer_id,
            name='favorite_lb0',
            project_id=self.project_id,
            vip_address=self.vip_address,
            vip_network_id=self.vip_network_id)
        mock.patch.object(
            ovn_helper.OvnProviderHelper, '_find_ovn_lbs',
            side_effect=lambda x, protocol=None:
                self.ovn_lb if protocol else [self.ovn_lb]).start()
        self.mock_find_lb_pool_key = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            '_find_ovn_lb_with_pool_key',
            return_value=self.ovn_lb).start()
        self.mock_find_ovn_lbs_with_retry = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            '_find_ovn_lbs_with_retry',
            return_value=self.ovn_lb).start()
        self.mock_get_subnet_from_pool = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            '_get_subnet_from_pool',
            return_value=(None, None)).start()
        self.mock_check_ip_in_subnet = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            '_check_ip_in_subnet',
            return_value=True).start()
        self.fake_fip = '1.2.3.4'
        self.lsp = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={
                'external_ids': {
                    ovn_const.OVN_PORT_FIP_EXT_ID_KEY: self.fake_fip,
                    ovn_const.OVN_PORT_NAME_EXT_ID_KEY:
                    'ovn-lb-vip-d571f37e-5708-48a1-8ba8-fc5d9ce36eac',
                },
                'type': 'localnet',
                'options': {},
            })
        self.lsp_addi = fakes.FakeOvsdbRow.create_one_ovsdb_row(
            attrs={
                'external_ids': {
                    ovn_const.OVN_PORT_FIP_EXT_ID_KEY: self.fake_fip,
                    ovn_const.OVN_PORT_NAME_EXT_ID_KEY:
                    'ovn-lb-vip-additional--\
                    19d4c20e-05ff-4534-9d5c-214f2979db40',
                },
                'type': 'localnet',
                'options': {},
            })
        self.mock_get_lsp = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            'get_lsp',
            return_value=self.lsp).start()
        self.mock_find_ovn_lbs_with_retry = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            '_find_ovn_lbs_with_retry',
            return_value=[self.ovn_lb]).start()

    def test_check_for_allowed_cidrs_exception(self):
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver._check_for_allowed_cidrs, '10.0.0.1')

    def test__ip_version_differs(self):
        self.assertFalse(self.driver._ip_version_differs(self.ref_member))
        self.ref_member.address = 'fc00::1'
        self.assertTrue(self.driver._ip_version_differs(self.ref_member))

    def test__ip_version_differs_lb_additional_vips(self):
        self.mock_find_ovn_lb_by_pool_id = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            '_find_ovn_lb_by_pool_id').start()
        self.mock_find_ovn_lb_by_pool_id.return_value = (_,
                                                         self.ovn_lb_addi_vips)
        self.ref_member.address = 'fc00::1'
        self.assertFalse(self.driver._ip_version_differs(self.ref_member))

    def test__ip_version_differs_lb_not_found(self):
        self.mock_find_ovn_lb_by_pool_id = mock.patch.object(
            ovn_helper.OvnProviderHelper,
            '_find_ovn_lb_by_pool_id').start()
        self.mock_find_ovn_lb_by_pool_id.return_value = (_, None)
        self.assertFalse(self.driver._ip_version_differs(self.ref_member))

    def test__ip_version_differs_pool_disabled(self):
        self.mock_find_lb_pool_key.side_effect = [None, self.ovn_lb]
        self.driver._ip_version_differs(self.ref_member)
        self.mock_find_lb_pool_key.assert_has_calls([
            mock.call('pool_%s' % self.pool_id),
            mock.call('pool_%s:D' % self.pool_id)])

    def _test_member_create(self, member):
        info = {'id': self.ref_member.member_id,
                'address': self.ref_member.address,
                'protocol_port': self.ref_member.protocol_port,
                'pool_id': self.ref_member.pool_id,
                'subnet_id': self.ref_member.subnet_id,
                'admin_state_up': self.ref_member.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_MEMBER_CREATE,
                         'info': info}
        info_dvr = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'action': ovn_const.REQ_INFO_MEMBER_ADDED}
        expected_dict_dvr = {
            'type': ovn_const.REQ_TYPE_HANDLE_MEMBER_DVR,
            'info': info_dvr}
        self.driver.member_create(member)
        expected = [
            mock.call(expected_dict),
            mock.call(expected_dict_dvr)]
        self.mock_add_request.assert_has_calls(expected)

    def test_member_create(self):
        self._test_member_create(self.ref_member)

    def test_member_create_failure(self):
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.fail_member)

    def test_member_create_different_ip_version(self):
        self.ref_member.address = 'fc00::1'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)

    def test_member_create_different_ip_version_lb_disable(self):
        self.driver._ovn_helper._find_ovn_lb_with_pool_key.side_effect = [
            None, self.ovn_lb]
        self.ref_member.address = 'fc00::1'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)
        self.driver._ovn_helper._find_ovn_lb_with_pool_key.assert_has_calls(
            [mock.call('pool_%s' % self.pool_id),
             mock.call('pool_%s%s' % (self.pool_id, ':D'))])

    def test_member_create_no_subnet_provided(self):
        self.ref_member.subnet_id = data_models.UnsetType()
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)
        self.ref_member.subnet_id = None
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)

    def test_member_create_no_subnet_provided_get_from_pool(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        self.driver._ovn_helper._check_ip_in_subnet.return_value = False
        self.ref_member.subnet_id = data_models.UnsetType()
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)
        self.ref_member.subnet_id = None
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)

    def test_member_create_no_subnet_provided_get_from_pool_failed(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        member = copy.copy(self.ref_member)
        member.subnet_id = data_models.UnsetType()
        self._test_member_create(member)
        member.subnet_id = None
        self._test_member_create(member)

    def test__check_monitor_options_member_no_monitor_data(self):
        self.ref_member.monitor_address = None
        self.assertFalse(self.driver._check_monitor_options(self.ref_member))

    def test_member_create_monitor_opts(self):
        self.ref_member.monitor_address = '172.20.20.1'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)
        self.ref_member.monitor_port = '80'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_create, self.ref_member)

    def test_member_create_no_set_admin_state_up(self):
        self.ref_member.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_member.member_id,
                'address': self.ref_member.address,
                'protocol_port': self.ref_member.protocol_port,
                'pool_id': self.ref_member.pool_id,
                'subnet_id': self.ref_member.subnet_id,
                'admin_state_up': True}
        expected_dict = {'type': ovn_const.REQ_TYPE_MEMBER_CREATE,
                         'info': info}
        expected_dict_dvr = {'type': ovn_const.REQ_TYPE_HANDLE_MEMBER_DVR,
                             'info': mock.ANY}
        expected = [
            mock.call(expected_dict),
            mock.call(expected_dict_dvr)]
        self.driver.member_create(self.ref_member)
        self.mock_add_request.assert_has_calls(expected)

    def test_member_update(self):
        info = {'id': self.update_member.member_id,
                'address': self.ref_member.address,
                'protocol_port': self.ref_member.protocol_port,
                'pool_id': self.ref_member.pool_id,
                'admin_state_up': self.update_member.admin_state_up,
                'old_admin_state_up': self.ref_member.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_MEMBER_UPDATE,
                         'info': info}
        self.driver.member_update(self.ref_member, self.update_member)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_member_update_missing_subnet_id(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        info = {'id': self.update_member.member_id,
                'address': self.ref_member.address,
                'protocol_port': self.ref_member.protocol_port,
                'pool_id': self.ref_member.pool_id,
                'admin_state_up': self.update_member.admin_state_up,
                'old_admin_state_up': self.ref_member.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_MEMBER_UPDATE,
                         'info': info}
        member = copy.copy(self.ref_member)
        member.subnet_id = data_models.UnsetType()
        self.driver.member_update(member, self.update_member)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_member_update_unset_admin_state_up(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        self.update_member.admin_state_up = data_models.UnsetType()
        info = {'id': self.update_member.member_id,
                'address': self.ref_member.address,
                'protocol_port': self.ref_member.protocol_port,
                'pool_id': self.ref_member.pool_id,
                'old_admin_state_up': self.ref_member.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_MEMBER_UPDATE,
                         'info': info}
        member = copy.copy(self.ref_member)
        member.subnet_id = data_models.UnsetType()
        self.driver.member_update(member, self.update_member)
        self.mock_add_request.assert_called_once_with(expected_dict)

    @mock.patch.object(ovn_driver.OvnProviderDriver, '_ip_version_differs')
    def test_member_update_no_ip_addr(self, mock_ip_differs):
        self.update_member.address = None
        self.driver.member_update(self.ref_member, self.update_member)
        mock_ip_differs.assert_not_called()

    def test_member_batch_update(self):
        self.driver.member_batch_update(self.pool_id,
                                        [self.ref_member, self.update_member])
        self.assertEqual(self.mock_add_request.call_count, 4)

    def test_member_batch_update_member_delete(self):
        info_md = {
            'id': self.ref_member.member_id,
            'address': mock.ANY,
            'protocol_port': mock.ANY,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id}
        expected_dict_md = {
            'type': ovn_const.REQ_TYPE_MEMBER_DELETE,
            'info': info_md}
        expected = [
            mock.call(expected_dict_md)]
        self.driver.member_batch_update(self.pool_id, [])
        self.assertEqual(self.mock_add_request.call_count, 2)
        self.mock_add_request.assert_has_calls(expected)

    def test_member_batch_update_no_members(self):
        pool_key = 'pool_%s' % self.pool_id
        ovn_lb = copy.copy(self.ovn_lb)
        ovn_lb.external_ids[pool_key] = []
        self.mock_find_lb_pool_key.return_value = ovn_lb
        self.driver.member_batch_update(self.pool_id,
                                        [self.ref_member, self.update_member])
        self.assertEqual(self.mock_add_request.call_count, 2)

    def test_member_batch_update_skipped_monitor(self):
        self.ref_member.monitor_address = '10.11.1.1'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_batch_update,
                          self.pool_id,
                          [self.ref_member])

    def test_member_batch_update_skipped_mixed_ip(self):
        self.ref_member.address = 'fc00::1'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_batch_update,
                          self.pool_id,
                          [self.ref_member])

    def test_member_batch_update_unset_admin_state_up(self):
        self.ref_member.admin_state_up = data_models.UnsetType()
        self.driver.member_batch_update(self.pool_id, [self.ref_member])
        self.assertEqual(self.mock_add_request.call_count, 3)

    def test_member_batch_update_toggle_admin_state_up(self):
        info_mu = {
            'id': self.ref_member.member_id,
            'address': self.member_address,
            'protocol_port': self.member_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'admin_state_up': False}
        expected_dict_mu = {
            'type': ovn_const.REQ_TYPE_MEMBER_UPDATE,
            'info': info_mu}
        expected = [
            mock.call(expected_dict_mu)]
        self.ref_member.admin_state_up = False
        self.ref_member.address = self.member_address
        self.ref_member.protocol_port = self.member_port
        self.driver.member_batch_update(self.pool_id, [self.ref_member])
        self.assertEqual(self.mock_add_request.call_count, 1)
        self.mock_add_request.assert_has_calls(expected)

    def test_member_batch_update_missing_subnet_id(self):
        self.ref_member.subnet_id = None
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_batch_update,
                          self.pool_id, [self.ref_member])

    def test_member_batch_update_missing_subnet_id_get_from_pool(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        self.ref_member.subnet_id = None
        self.driver.member_batch_update(self.pool_id, [self.ref_member])

    def test_member_batch_update_missing_subnet_id_get_from_pool_fail(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        self.driver._ovn_helper._check_ip_in_subnet.return_value = False
        self.ref_member.subnet_id = None
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_batch_update,
                          self.pool_id, [self.ref_member])

    def test_member_update_failure(self):
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_update, self.ref_member,
                          self.fail_member)

    def test_member_update_different_ip_version(self):
        self.ref_member.address = 'fc00::1'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_update, self.ref_member,
                          self.ref_member)

    def test_member_delete(self):
        info = {'id': self.ref_member.member_id,
                'address': self.ref_member.address,
                'protocol_port': self.ref_member.protocol_port,
                'pool_id': self.ref_member.pool_id,
                'subnet_id': self.ref_member.subnet_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_MEMBER_DELETE,
                         'info': info}
        info_dvr = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'action': ovn_const.REQ_INFO_MEMBER_DELETED}
        expected_dict_dvr = {
            'type': ovn_const.REQ_TYPE_HANDLE_MEMBER_DVR,
            'info': info_dvr}
        self.driver.member_delete(self.ref_member)
        expected = [
            mock.call(expected_dict),
            mock.call(expected_dict_dvr)]
        self.mock_add_request.assert_has_calls(expected)

    def test_member_delete_missing_subnet_id(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        info = {'id': self.ref_member.member_id,
                'address': self.ref_member.address,
                'protocol_port': self.ref_member.protocol_port,
                'pool_id': self.ref_member.pool_id,
                'subnet_id': self.ref_member.subnet_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_MEMBER_DELETE,
                         'info': info}
        info_dvr = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'action': ovn_const.REQ_INFO_MEMBER_DELETED}
        expected_dict_dvr = {
            'type': ovn_const.REQ_TYPE_HANDLE_MEMBER_DVR,
            'info': info_dvr}

        member = copy.copy(self.ref_member)
        member.subnet_id = data_models.UnsetType()
        self.driver.member_delete(member)
        expected = [
            mock.call(expected_dict),
            mock.call(expected_dict_dvr)]
        self.mock_add_request.assert_has_calls(expected)

    def test_member_delete_missing_subnet_id_differs_from_lb_vip(self):
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            self.ref_member.subnet_id, '198.52.100.0/24')
        self.driver._ovn_helper._check_ip_in_subnet.return_value = False
        self.ref_member.subnet_id = data_models.UnsetType()
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.member_delete, self.ref_member)

    def test_listener_create(self):
        info = {'id': self.ref_listener.listener_id,
                'protocol': self.ref_listener.protocol,
                'protocol_port': self.ref_listener.protocol_port,
                'default_pool_id': self.ref_listener.default_pool_id,
                'admin_state_up': self.ref_listener.admin_state_up,
                'loadbalancer_id': self.ref_listener.loadbalancer_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_CREATE,
                         'info': info}
        self.driver.listener_create(self.ref_listener)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_listener_create_unset_admin_state_up(self):
        self.ref_listener.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_listener.listener_id,
                'protocol': self.ref_listener.protocol,
                'protocol_port': self.ref_listener.protocol_port,
                'default_pool_id': self.ref_listener.default_pool_id,
                'admin_state_up': True,
                'loadbalancer_id': self.ref_listener.loadbalancer_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_CREATE,
                         'info': info}
        self.driver.listener_create(self.ref_listener)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_listener_create_unsupported_protocol(self):
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.listener_create, self.fail_listener)

    def test_listener_create_multiple_protocols(self):
        self.ovn_lb.protocol = ['TCP']
        info = {'id': self.ref_listener.listener_id,
                'protocol': self.ref_listener.protocol,
                'protocol_port': self.ref_listener.protocol_port,
                'default_pool_id': self.ref_listener.default_pool_id,
                'admin_state_up': self.ref_listener.admin_state_up,
                'loadbalancer_id': self.ref_listener.loadbalancer_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_CREATE,
                         'info': info}
        self.driver.listener_create(self.ref_listener)
        self.mock_add_request.assert_called_once_with(expected_dict)
        self.ovn_lb.protocol = ['UDP']
        info['protocol'] = 'UDP'
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_CREATE,
                         'info': info}
        self.driver.listener_create(self.ref_listener)
        self.ovn_lb.protocol = ['SCTP']
        info['protocol'] = 'SCTP'
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_CREATE,
                         'info': info}
        self.driver.listener_create(self.ref_listener)

    def test_listener_update(self):
        info = {'id': self.ref_listener.listener_id,
                'protocol_port': self.ref_listener.protocol_port,
                'protocol': self.ref_pool.protocol,
                'admin_state_up': self.ref_listener.admin_state_up,
                'loadbalancer_id': self.ref_listener.loadbalancer_id}
        if self.ref_listener.default_pool_id:
            info['default_pool_id'] = self.ref_listener.default_pool_id
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_UPDATE,
                         'info': info}
        self.driver.listener_update(self.ref_listener, self.ref_listener)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_listener_update_unset_admin_state_up(self):
        self.ref_listener.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_listener.listener_id,
                'protocol_port': self.ref_listener.protocol_port,
                'protocol': self.ref_pool.protocol,
                'loadbalancer_id': self.ref_listener.loadbalancer_id}
        if self.ref_listener.default_pool_id:
            info['default_pool_id'] = self.ref_listener.default_pool_id
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_UPDATE,
                         'info': info}
        self.driver.listener_update(self.ref_listener, self.ref_listener)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_listener_update_unset_default_pool_id(self):
        self.ref_listener.default_pool_id = data_models.UnsetType()
        info = {'id': self.ref_listener.listener_id,
                'protocol_port': self.ref_listener.protocol_port,
                'protocol': self.ref_pool.protocol,
                'admin_state_up': self.ref_listener.admin_state_up,
                'loadbalancer_id': self.ref_listener.loadbalancer_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_UPDATE,
                         'info': info}
        self.driver.listener_update(self.ref_listener, self.ref_listener)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_listener_delete(self):
        info = {'id': self.ref_listener.listener_id,
                'protocol_port': self.ref_listener.protocol_port,
                'protocol': self.ref_pool.protocol,
                'loadbalancer_id': self.ref_listener.loadbalancer_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_LISTENER_DELETE,
                         'info': info}
        self.driver.listener_delete(self.ref_listener)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_loadbalancer_fully_populate_create(self):
        info = {
            'id': self.ref_lb_fully_populated.loadbalancer_id,
            'vip_address': self.ref_lb_fully_populated.vip_address,
            'vip_network_id': self.ref_lb_fully_populated.vip_network_id,
            'admin_state_up': self.ref_lb_fully_populated.admin_state_up}
        info_listener = {
            'id': self.ref_listener.listener_id,
            'protocol': self.ref_listener.protocol,
            'protocol_port': self.ref_listener.protocol_port,
            'default_pool_id': self.ref_listener.default_pool_id,
            'admin_state_up': self.ref_listener.admin_state_up,
            'loadbalancer_id': self.ref_listener.loadbalancer_id}
        info_pool = {
            'id': self.ref_pool.pool_id,
            'loadbalancer_id': self.ref_pool.loadbalancer_id,
            'listener_id': self.ref_pool.listener_id,
            'protocol': self.ref_pool.protocol,
            'lb_algorithm': constants.LB_ALGORITHM_SOURCE_IP_PORT,
            'admin_state_up': self.ref_pool.admin_state_up,
            'session_persistence': {'type': 'SOURCE_IP'}}
        info_member = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'protocol_port': self.ref_member.protocol_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'admin_state_up': self.ref_member.admin_state_up}
        info_dvr = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'action': ovn_const.REQ_INFO_MEMBER_ADDED}
        expected_lb_dict = {
            'type': ovn_const.REQ_TYPE_LB_CREATE,
            'info': info}
        expected_listener_dict = {
            'type': ovn_const.REQ_TYPE_LISTENER_CREATE,
            'info': info_listener}
        expected_pool_dict = {
            'type': ovn_const.REQ_TYPE_POOL_CREATE,
            'info': info_pool}
        expected_member_dict = {
            'type': ovn_const.REQ_TYPE_MEMBER_CREATE,
            'info': info_member}
        expected_dict_dvr = {
            'type': ovn_const.REQ_TYPE_HANDLE_MEMBER_DVR,
            'info': info_dvr}
        calls = [mock.call(expected_lb_dict),
                 mock.call(expected_listener_dict),
                 mock.call(expected_pool_dict),
                 mock.call(expected_member_dict),
                 mock.call(expected_dict_dvr)]
        self.driver.loadbalancer_create(self.ref_lb_fully_populated)
        self.mock_add_request.assert_has_calls(calls)

    def test_loadbalancer_create(self):
        info = {'id': self.ref_lb0.loadbalancer_id,
                'vip_address': self.ref_lb0.vip_address,
                'vip_network_id': self.ref_lb0.vip_network_id,
                'admin_state_up': self.ref_lb0.admin_state_up}
        expected_dict = {
            'type': ovn_const.REQ_TYPE_LB_CREATE,
            'info': info}
        calls = [mock.call(expected_dict)]
        self.driver.loadbalancer_create(self.ref_lb0)
        self.mock_add_request.assert_has_calls(calls)

    @mock.patch.object(ovn_helper.OvnProviderHelper, 'vip_port_update_handler')
    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_basic(self, net_cli, m_vpu):
        net_cli.return_value.ips.return_value = [
            mock.Mock(floating_ip_address=self.fake_fip)]
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'
        self.driver._fip_sync(self.ref_lb_fully_sync_populated)
        m_vpu.assert_called_once_with(
            vip_lp=self.lsp, fip=self.fake_fip,
            action=ovn_const.REQ_INFO_ACTION_SYNC)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync(self, net_cli):
        net_cli.return_value.ips.return_value = [
            mock.Mock(floating_ip_address=self.fake_fip)]
        fake_port = fakes.FakePort.create_one_port()

        fake_port['fixed_ips'][0]['ip_address'] = self.fake_vip

        net_cli.return_value.get_port.return_value = fake_port
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'

        with mock.patch.object(ovn_helper.OvnProviderHelper,
                               'handle_vip_fip') as mock_handle_vip_fip:
            info = {
                'ovn_lb': self.ovn_lb,
                'vip_fip': self.fake_fip,
                'vip_related': [self.fake_vip],
                'additional_vip_fip': False,
                'action': ovn_const.REQ_INFO_ACTION_SYNC
            }
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            mock_handle_vip_fip.assert_called_once_with(info)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_ovn_lb_not_found(self, net_cli):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        net_cli.return_value.ips.return_value = [
            mock.Mock(floating_ip_address=self.fake_fip)]
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'
        with mock.patch.object(ovn_helper, 'LOG') as m_l:
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            m_l.debug.assert_called()
        self.mock_add_request.assert_not_called()

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_addi(self, net_cli):
        net_cli.return_value.ips.return_value = [
            mock.Mock(floating_ip_address=self.fake_fip)]
        fake_port = fakes.FakePort.create_one_port()
        self.mock_get_lsp.return_value = self.lsp_addi
        fake_port['fixed_ips'][0]['ip_address'] = self.fake_vip

        net_cli.return_value.get_port.return_value = fake_port
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'

        with mock.patch.object(ovn_helper.OvnProviderHelper,
                               'handle_vip_fip') as mock_handle_vip_fip:
            info = {
                'ovn_lb': self.ovn_lb,
                'vip_fip': self.fake_fip,
                'vip_related': [self.fake_vip],
                'additional_vip_fip': True,
                'action': ovn_const.REQ_INFO_ACTION_SYNC
            }
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            mock_handle_vip_fip.assert_called_once_with(info)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_os_err(self, net_cli):
        net_cli.return_value.ips.side_effect = [
            openstack.exceptions.HttpException]
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'

        calls = [
            mock.call('Floating IP not found for loadbalancer '
                      f'{self.ref_lb_fully_sync_populated.loadbalancer_id}'),
            mock.call('Floating IP not consistent between Logic Switch Port '
                      f'and Neutron. Found FIP {self.fake_fip} configured in '
                      f'LSP {self.lsp.name}, but no FIP configured from '
                      'Neutron. Please run command `neutron-ovn-db-sync-util` '
                      'first to sync OVN DB with Neutron DB.')]

        with mock.patch.object(ovn_driver, 'LOG') as m_l:
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            m_l.warn.assert_has_calls(calls)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_lsp_mismatch(self, net_cli):
        net_cli.return_value.ips.return_value = [
            mock.Mock(floating_ip_address='1.2.3.5')]
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'
        msg = ('Floating IP not consistent between Logic Switch Port and '
               f'Neutron. Found FIP {self.fake_fip} in LSP {self.lsp.name}, '
               'but we have 1.2.3.5 from Neutron. Skip sync FIP for '
               'loadbalancer '
               f'{self.ref_lb_fully_sync_populated.loadbalancer_id}. '
               'Please run command `neutron-ovn-db-sync-util` first to '
               'sync OVN DB with Neutron DB.')

        with mock.patch.object(ovn_driver, 'LOG') as m_l:
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            m_l.warn.assert_called_once_with(msg)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_lsp_not_found(self, net_cli):
        net_cli.return_value.ips.return_value = [
            mock.Mock(floating_ip_address='1.2.3.4')]
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'

        self.mock_get_lsp.return_value = None
        msg = (
            'Logic Switch Port not found for port foo. Skip sync FIP for '
            'loadbalancer '
            f'{self.ref_lb_fully_sync_populated.loadbalancer_id}. Please '
            'run command `neutron-ovn-db-sync-util` first to sync OVN DB '
            'with Neutron DB.')

        with mock.patch.object(ovn_driver, 'LOG') as m_l:
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            m_l.warn.assert_called_once_with(msg)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_no_neutron_fip(self, net_cli):
        net_cli.return_value.ips.return_value = []
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'

        calls = [
            mock.call('Floating IP not found for loadbalancer '
                      f'{self.ref_lb_fully_sync_populated.loadbalancer_id}'),
            mock.call('Floating IP not consistent between Logic Switch Port '
                      f'and Neutron. Found FIP {self.fake_fip} configured in '
                      f'LSP {self.lsp.name}, but no FIP configured from '
                      'Neutron. Please run command `neutron-ovn-db-sync-util` '
                      'first to sync OVN DB with Neutron DB.')]
        self.driver.pool_create(self.ref_pool)
        self.mock_add_request
        with mock.patch.object(ovn_driver, 'LOG') as m_l:
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            m_l.warn.assert_has_calls(calls)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_no_neutron_fip_no_lsp(self, net_cli):
        net_cli.return_value.ips.return_value = []
        self.ref_lb_fully_sync_populated.vip_port_id = 'foo'
        self.mock_get_lsp.return_value = None
        self.driver._fip_sync(self.ref_lb_fully_sync_populated)

    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_fip_sync_no_port_id(self, net_cli):
        net_cli.return_value.ips.return_value = [
            mock.Mock(floating_ip_address='1.2.3.4')]

        msg = ('VIP Port or Network not set for loadbalancer '
               f'{self.ref_lb_fully_sync_populated.loadbalancer_id}, '
               'skip FIP sync.')
        with mock.patch.object(ovn_driver, 'LOG') as m_l:
            self.driver._fip_sync(self.ref_lb_fully_sync_populated)
            m_l.debug.assert_called_once_with(msg)

    def test_loadbalancer_create_additional_vips(self):
        info = {'id': self.ref_lb2.loadbalancer_id,
                'vip_address': self.ref_lb2.vip_address,
                'vip_network_id': self.ref_lb2.vip_network_id,
                'additional_vips': self.ref_lb2.additional_vips,
                'admin_state_up': self.ref_lb2.admin_state_up}
        expected_dict = {
            'type': ovn_const.REQ_TYPE_LB_CREATE,
            'info': info}
        calls = [mock.call(expected_dict)]
        self.driver.loadbalancer_create(self.ref_lb2)
        self.mock_add_request.assert_has_calls(calls)

    def test_loadbalancer_create_member_without_subnet_id(self):
        self.ref_member.subnet_id = data_models.UnsetType()
        info = {
            'id': self.ref_lb_fully_populated.loadbalancer_id,
            'vip_address': self.ref_lb_fully_populated.vip_address,
            'vip_network_id': self.ref_lb_fully_populated.vip_network_id,
            'admin_state_up': self.ref_lb_fully_populated.admin_state_up}
        info_listener = {
            'id': self.ref_listener.listener_id,
            'protocol': self.ref_listener.protocol,
            'protocol_port': self.ref_listener.protocol_port,
            'default_pool_id': self.ref_listener.default_pool_id,
            'admin_state_up': self.ref_listener.admin_state_up,
            'loadbalancer_id': self.ref_listener.loadbalancer_id}
        info_pool = {
            'id': self.ref_pool.pool_id,
            'loadbalancer_id': self.ref_pool.loadbalancer_id,
            'listener_id': self.ref_pool.listener_id,
            'protocol': self.ref_pool.protocol,
            'lb_algorithm': constants.LB_ALGORITHM_SOURCE_IP_PORT,
            'admin_state_up': self.ref_pool.admin_state_up,
            'session_persistence': {'type': 'SOURCE_IP'}}
        info_member = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'protocol_port': self.ref_member.protocol_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_lb_fully_populated.vip_subnet_id,
            'admin_state_up': self.ref_member.admin_state_up}
        info_dvr = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_lb_fully_populated.vip_subnet_id,
            'action': ovn_const.REQ_INFO_MEMBER_ADDED}
        expected_lb_dict = {
            'type': ovn_const.REQ_TYPE_LB_CREATE,
            'info': info}
        expected_listener_dict = {
            'type': ovn_const.REQ_TYPE_LISTENER_CREATE,
            'info': info_listener}
        expected_pool_dict = {
            'type': ovn_const.REQ_TYPE_POOL_CREATE,
            'info': info_pool}
        expected_member_dict = {
            'type': ovn_const.REQ_TYPE_MEMBER_CREATE,
            'info': info_member}
        expected_dict_dvr = {
            'type': ovn_const.REQ_TYPE_HANDLE_MEMBER_DVR,
            'info': info_dvr}
        calls = [mock.call(expected_lb_dict),
                 mock.call(expected_listener_dict),
                 mock.call(expected_pool_dict),
                 mock.call(expected_member_dict),
                 mock.call(expected_dict_dvr)]
        self.driver.loadbalancer_create(self.ref_lb_fully_populated)
        self.mock_add_request.assert_has_calls(calls)

    def test_loadbalancer_create_unset_listeners(self):
        self.ref_lb0.listeners = data_models.UnsetType()
        info = {'id': self.ref_lb0.loadbalancer_id,
                'vip_address': self.ref_lb0.vip_address,
                'vip_network_id': self.ref_lb0.vip_network_id,
                'admin_state_up': False}
        expected_dict = {
            'type': ovn_const.REQ_TYPE_LB_CREATE,
            'info': info}
        calls = [mock.call(expected_dict)]
        self.driver.loadbalancer_create(self.ref_lb0)
        self.mock_add_request.assert_has_calls(calls)

    def test_loadbalancer_create_unset_admin_state_up(self):
        self.ref_lb0.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_lb0.loadbalancer_id,
                'vip_address': self.ref_lb0.vip_address,
                'vip_network_id': self.ref_lb0.vip_network_id,
                'admin_state_up': True}
        expected_dict = {
            'type': ovn_const.REQ_TYPE_LB_CREATE,
            'info': info}
        calls = [mock.call(expected_dict)]
        self.driver.loadbalancer_create(self.ref_lb0)
        self.mock_add_request.assert_has_calls(calls)

    def test_loadbalancer_update(self):
        info = {'id': self.ref_lb1.loadbalancer_id,
                'admin_state_up': self.ref_lb1.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_LB_UPDATE,
                         'info': info}
        self.driver.loadbalancer_update(self.ref_lb0, self.ref_lb1)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_loadbalancer_update_unset_admin_state_up(self):
        self.ref_lb1.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_lb1.loadbalancer_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_LB_UPDATE,
                         'info': info}
        self.driver.loadbalancer_update(self.ref_lb0, self.ref_lb1)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_loadbalancer_delete(self):
        info = {'id': self.ref_lb0.loadbalancer_id,
                'cascade': False}
        expected_dict = {'type': ovn_const.REQ_TYPE_LB_DELETE,
                         'info': info}
        self.driver.loadbalancer_delete(self.ref_lb1)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_loadbalancer_failover(self):
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.loadbalancer_failover,
                          self.ref_lb0.loadbalancer_id)

    def test_pool_create_unsupported_protocol(self):
        self.ref_pool.protocol = 'HTTP'
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.pool_create, self.ref_pool)

    def test_pool_create_leastcount_algo(self):
        self.ref_pool.lb_algorithm = constants.LB_ALGORITHM_LEAST_CONNECTIONS
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.pool_create, self.ref_pool)

    def test_pool_create(self):
        info = {'id': self.ref_pool.pool_id,
                'loadbalancer_id': self.ref_pool.loadbalancer_id,
                'listener_id': self.ref_pool.listener_id,
                'protocol': self.ref_pool.protocol,
                'lb_algorithm': constants.LB_ALGORITHM_SOURCE_IP_PORT,
                'admin_state_up': self.ref_pool.admin_state_up,
                'session_persistence': {'type': 'SOURCE_IP'}}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_CREATE,
                         'info': info}
        self.driver.pool_create(self.ref_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_pool_create_with_health_monitor(self):
        self.ref_pool.healthmonitor = self.ref_health_monitor
        info = {'id': self.ref_pool.pool_id,
                'loadbalancer_id': self.ref_pool.loadbalancer_id,
                'listener_id': self.ref_pool.listener_id,
                'protocol': self.ref_pool.protocol,
                'lb_algorithm': constants.LB_ALGORITHM_SOURCE_IP_PORT,
                'admin_state_up': self.ref_pool.admin_state_up,
                'session_persistence': {'type': 'SOURCE_IP'}}
        info_hm = {'id': self.ref_health_monitor.healthmonitor_id,
                   'pool_id': self.ref_health_monitor.pool_id,
                   'type': self.ref_health_monitor.type,
                   'interval': self.ref_health_monitor.delay,
                   'timeout': self.ref_health_monitor.timeout,
                   'failure_count': self.ref_health_monitor.max_retries_down,
                   'success_count': self.ref_health_monitor.max_retries,
                   'admin_state_up': self.ref_health_monitor.admin_state_up}

        expected_pool_dict = {'type': ovn_const.REQ_TYPE_POOL_CREATE,
                              'info': info}
        expected_hm_dict = {'type': ovn_const.REQ_TYPE_HM_CREATE,
                            'info': info_hm}
        calls = [mock.call(expected_pool_dict),
                 mock.call(expected_hm_dict)]
        self.driver.pool_create(self.ref_pool)
        self.mock_add_request.assert_has_calls(calls)

    def test_pool_create_unset_admin_state_up(self):
        self.ref_pool.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_pool.pool_id,
                'loadbalancer_id': self.ref_pool.loadbalancer_id,
                'protocol': self.ref_pool.protocol,
                'lb_algorithm': constants.LB_ALGORITHM_SOURCE_IP_PORT,
                'listener_id': self.ref_pool.listener_id,
                'admin_state_up': True,
                'session_persistence': {'type': 'SOURCE_IP'}}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_CREATE,
                         'info': info}
        self.driver.pool_create(self.ref_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_pool_create_unset_session_persistence(self):
        self.ref_pool.session_persistence = data_models.UnsetType()
        info = {'id': self.ref_pool.pool_id,
                'loadbalancer_id': self.ref_pool.loadbalancer_id,
                'protocol': self.ref_pool.protocol,
                'lb_algorithm': constants.LB_ALGORITHM_SOURCE_IP_PORT,
                'listener_id': self.ref_pool.listener_id,
                'admin_state_up': self.ref_pool.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_CREATE,
                         'info': info}
        self.driver.pool_create(self.ref_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_pool_delete(self):
        # Pretent we don't have members
        self.ref_pool.members = []
        info = {'id': self.ref_pool.pool_id,
                'protocol': self.ref_pool.protocol,
                'loadbalancer_id': self.ref_pool.loadbalancer_id}
        expected = {'type': ovn_const.REQ_TYPE_POOL_DELETE,
                    'info': info}
        self.driver.pool_delete(self.ref_pool)
        self.mock_add_request.assert_called_once_with(expected)

    def test_pool_delete_with_members_and_hm(self):
        self.ref_pool.healthmonitor = self.ref_health_monitor
        info = {'id': self.ref_pool.pool_id,
                'protocol': self.ref_pool.protocol,
                'loadbalancer_id': self.ref_pool.loadbalancer_id}
        expected = {'type': ovn_const.REQ_TYPE_POOL_DELETE,
                    'info': info}
        info_hm = {'id': self.ref_pool.healthmonitor.healthmonitor_id,
                   'pool_id': self.ref_pool.pool_id}
        info_member = {'id': self.ref_member.member_id,
                       'pool_id': self.ref_member.pool_id,
                       'subnet_id': self.ref_member.subnet_id,
                       'protocol_port': self.ref_member.protocol_port,
                       'address': self.ref_member.address}
        expected_hm = {
            'type': ovn_const.REQ_TYPE_HM_DELETE,
            'info': info_hm}
        expected_members = {
            'type': ovn_const.REQ_TYPE_MEMBER_DELETE,
            'info': info_member}
        expected_members_dvr = {
            'type': ovn_const.REQ_TYPE_HANDLE_MEMBER_DVR,
            'info': mock.ANY}
        calls = [mock.call(expected_hm),
                 mock.call(expected_members),
                 mock.call(expected_members_dvr),
                 mock.call(expected)]
        self.driver.pool_delete(self.ref_pool)
        self.mock_add_request.assert_has_calls(calls)

    def test_pool_update(self):
        info = {'id': self.ref_update_pool.pool_id,
                'loadbalancer_id': self.ref_update_pool.loadbalancer_id,
                'protocol': self.ref_pool.protocol,
                'admin_state_up': self.ref_update_pool.admin_state_up,
                'session_persistence': {'type': 'SOURCE_IP'}}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_UPDATE,
                         'info': info}
        self.driver.pool_update(self.ref_pool, self.ref_update_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_pool_update_unset_new_protocol(self):
        self.ref_update_pool.protocol = data_models.UnsetType()
        info = {'id': self.ref_update_pool.pool_id,
                'loadbalancer_id': self.ref_update_pool.loadbalancer_id,
                'protocol': self.ref_pool.protocol,
                'admin_state_up': self.ref_update_pool.admin_state_up,
                'session_persistence': {'type': 'SOURCE_IP'}}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_UPDATE,
                         'info': info}
        self.driver.pool_update(self.ref_pool, self.ref_update_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_pool_update_unset_new_lb_algorithm(self):
        self.ref_update_pool.lb_algorithm = data_models.UnsetType()
        info = {'id': self.ref_update_pool.pool_id,
                'loadbalancer_id': self.ref_update_pool.loadbalancer_id,
                'protocol': self.ref_pool.protocol,
                'admin_state_up': self.ref_update_pool.admin_state_up,
                'session_persistence': {'type': 'SOURCE_IP'}}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_UPDATE,
                         'info': info}
        self.driver.pool_update(self.ref_pool, self.ref_update_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_pool_update_unset_new_admin_state_up(self):
        self.ref_update_pool.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_update_pool.pool_id,
                'loadbalancer_id': self.ref_update_pool.loadbalancer_id,
                'protocol': self.ref_pool.protocol,
                'session_persistence': {'type': 'SOURCE_IP'}}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_UPDATE,
                         'info': info}
        self.driver.pool_update(self.ref_pool, self.ref_update_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_pool_update_unset_new_session_timeout(self):
        self.ref_update_pool.session_persistence = data_models.UnsetType()
        info = {'id': self.ref_update_pool.pool_id,
                'loadbalancer_id': self.ref_update_pool.loadbalancer_id,
                'protocol': self.ref_pool.protocol,
                'admin_state_up': self.ref_update_pool.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_POOL_UPDATE,
                         'info': info}
        self.driver.pool_update(self.ref_pool, self.ref_update_pool)
        self.mock_add_request.assert_called_once_with(expected_dict)

    @mock.patch.object(ovn_driver.OvnProviderDriver,
                       '_check_member_monitor_options')
    @mock.patch.object(ovn_driver.OvnProviderDriver,
                       '_ip_version_differs')
    def test_get_member_request_info_create_valid(
            self, mock_ip_version_differs, mock_check_monitor):
        mock_ip_version_differs.return_value = False
        result = self.driver._get_member_request_info(
            self.ref_member, create=True)

        mock_check_monitor.assert_called_once_with(self.ref_member)
        mock_ip_version_differs.assert_called_once_with(self.ref_member)
        expected_result = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'protocol_port': self.ref_member.protocol_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'admin_state_up': self.ref_member.admin_state_up,
        }

        self.assertEqual(result, expected_result)

    @mock.patch.object(ovn_driver.OvnProviderDriver, '_validate_hm_support')
    def test_get_healthmonitor_request_info_with_admin_state_up(
            self, mock_validate):
        result = self.driver._get_healthmonitor_request_info(
            self.ref_health_monitor)

        expected = {
            'id': self.ref_health_monitor.healthmonitor_id,
            'pool_id': self.ref_health_monitor.pool_id,
            'type': self.ref_health_monitor.type,
            'interval': self.ref_health_monitor.delay,
            'timeout': self.ref_health_monitor.timeout,
            'failure_count': self.ref_health_monitor.max_retries_down,
            'success_count': self.ref_health_monitor.max_retries,
            'admin_state_up': self.ref_health_monitor.admin_state_up
        }
        self.assertEqual(result, expected)
        mock_validate.assert_called_once_with(self.ref_health_monitor)

    @mock.patch.object(ovn_driver.OvnProviderDriver, '_validate_hm_support')
    def test_get_healthmonitor_request_info_with_unset_admin_state_up(
            self, mock_validate):
        self.ref_health_monitor.admin_state_up = data_models.UnsetType()
        result = self.driver._get_healthmonitor_request_info(
            self.ref_health_monitor)

        expected = {
            'id': self.ref_health_monitor.healthmonitor_id,
            'pool_id': self.ref_health_monitor.pool_id,
            'type': self.ref_health_monitor.type,
            'interval': self.ref_health_monitor.delay,
            'timeout': self.ref_health_monitor.timeout,
            'failure_count': self.ref_health_monitor.max_retries_down,
            'success_count': self.ref_health_monitor.max_retries,
            'admin_state_up': True
        }
        self.assertEqual(result, expected)
        mock_validate.assert_called_once_with(self.ref_health_monitor)

    @mock.patch.object(ovn_driver.OvnProviderDriver, '_validate_hm_support')
    def test_get_healthmonitor_request_info_with_admin_state_down(
            self, mock_validate):
        self.ref_health_monitor.admin_state_up = False
        result = self.driver._get_healthmonitor_request_info(
            self.ref_health_monitor)

        expected = {
            'id': self.ref_health_monitor.healthmonitor_id,
            'pool_id': self.ref_health_monitor.pool_id,
            'type': self.ref_health_monitor.type,
            'interval': self.ref_health_monitor.delay,
            'timeout': self.ref_health_monitor.timeout,
            'failure_count': self.ref_health_monitor.max_retries_down,
            'success_count': self.ref_health_monitor.max_retries,
            'admin_state_up': self.ref_health_monitor.admin_state_up
        }
        self.assertEqual(result, expected)
        mock_validate.assert_called_once_with(self.ref_health_monitor)

    @mock.patch.object(ovn_driver.OvnProviderDriver,
                       '_check_member_monitor_options')
    @mock.patch.object(ovn_driver.OvnProviderDriver,
                       '_ip_version_differs')
    def test_ip_version_differs_raises_exception(
            self, mock_ip_version_differs, mock_check_monitor):
        mock_ip_version_differs.return_value = True

        self.assertRaises(ovn_exc.IPVersionsMixingNotSupportedError,
                          self.driver._get_member_request_info,
                          self.ref_member,
                          create=True)

        mock_check_monitor.assert_called_once_with(self.ref_member)
        mock_ip_version_differs.assert_called_once_with(self.ref_member)

    def test_get_subnet_from_pool(self):
        self.ref_member.subnet_id = None
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            'subnet_2', '10.0.0.0/24')
        self.driver._ovn_helper._check_ip_in_subnet.return_value = True

        result = self.driver._get_member_request_info(self.ref_member,
                                                      create=True)

        self.driver._ovn_helper._get_subnet_from_pool.assert_called_once_with(
            self.ref_member.pool_id)
        self.driver._ovn_helper._check_ip_in_subnet.assert_called_once_with(
            self.ref_member.address, '10.0.0.0/24')

        expected_result = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'protocol_port': self.ref_member.protocol_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': 'subnet_2',
            'admin_state_up': self.ref_member.admin_state_up,
        }

        self.assertEqual(result, expected_result)

    def test_get_subnet_fails_raises_exception(self):
        self.ref_member.subnet_id = None
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (None,
                                                                      None)

        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver._get_member_request_info,
                          self.ref_member, create=True)

        self.driver._ovn_helper._get_subnet_from_pool.assert_called_once_with(
            self.ref_member.pool_id)

    def test_get_member_request_info_with_unset_type_admin_state_up(self):
        self.ref_member.admin_state_up = data_models.UnsetType()
        result = self.driver._get_member_request_info(self.ref_member,
                                                      create=True)

        expected_result = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'protocol_port': self.ref_member.protocol_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id,
            'admin_state_up': True,
        }
        self.assertEqual(result, expected_result)

    def test_get_member_request_info_create_false(self):
        result = self.driver._get_member_request_info(self.ref_member,
                                                      create=False)

        expected_result = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'protocol_port': self.ref_member.protocol_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': self.ref_member.subnet_id
        }
        self.assertEqual(result, expected_result)

    def test_get_member_request_info_with_unset_type_subnet(self):
        self.ref_member.subnet_id = data_models.UnsetType()
        self.driver._ovn_helper._get_subnet_from_pool.return_value = (
            'subnet_2', '10.0.0.0/24')
        self.driver._ovn_helper._check_ip_in_subnet.return_value = True

        result = self.driver._get_member_request_info(self.ref_member,
                                                      create=True)

        expected_result = {
            'id': self.ref_member.member_id,
            'address': self.ref_member.address,
            'protocol_port': self.ref_member.protocol_port,
            'pool_id': self.ref_member.pool_id,
            'subnet_id': 'subnet_2',
            'admin_state_up': self.ref_member.admin_state_up,
        }
        self.assertEqual(result, expected_result)

    def test_create_vip_port(self):
        with mock.patch.object(clients, 'get_neutron_client'):
            port_dict, add_vip_dicts = (
                self.driver.create_vip_port(self.loadbalancer_id,
                                            self.project_id,
                                            self.vip_dict, []))
            self.assertIsNotNone(port_dict.pop('vip_address', None))
            self.assertIsNotNone(port_dict.pop('vip_port_id', None))
            self.assertEqual(len(add_vip_dicts), 0)
            # The network_driver function is mocked, therefore the
            # created port vip_address and vip_port_id are also mocked.
            # Check if it exists and move on.
            # The finally output is include vip_address, vip_port_id,
            # vip_network_id and vip_subnet_id.
            for key, value in port_dict.items():
                self.assertEqual(value, self.vip_output[key])

    def test_create_vip_port_additional_vips(self):
        with mock.patch.object(clients, 'get_neutron_client'):
            port_dict, add_vip_dicts = (
                self.driver.create_vip_port(self.loadbalancer_id,
                                            self.project_id,
                                            self.vip_dict,
                                            self.additional_vips))
            self.assertIsNotNone(port_dict.pop('vip_address', None))
            self.assertIsNotNone(port_dict.pop('vip_port_id', None))
            self.assertIsNotNone(add_vip_dicts)
            self.assertEqual(len(add_vip_dicts), 1)
            # The network_driver function is mocked, therefore the
            # created port vip_address and vip_port_id are also mocked.
            # Check if it exists and move on.
            # The finally output is include vip_address, vip_port_id,
            # vip_network_id and vip_subnet_id.
            for key, value in port_dict.items():
                self.assertEqual(value, self.vip_output[key])

    def test_create_vip_port_exception(self):
        with mock.patch.object(clients, 'get_neutron_client',
                               side_effect=[RuntimeError]):
            self.assertRaises(
                exceptions.DriverError,
                self.driver.create_vip_port,
                self.loadbalancer_id,
                self.project_id,
                self.vip_dict,
                [])

    def test_health_monitor_create(self):
        info = {'id': self.ref_health_monitor.healthmonitor_id,
                'pool_id': self.ref_health_monitor.pool_id,
                'type': self.ref_health_monitor.type,
                'interval': self.ref_health_monitor.delay,
                'timeout': self.ref_health_monitor.timeout,
                'failure_count': self.ref_health_monitor.max_retries_down,
                'success_count': self.ref_health_monitor.max_retries,
                'admin_state_up': self.ref_health_monitor.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_HM_CREATE,
                         'info': info}
        self.driver.health_monitor_create(self.ref_health_monitor)
        self.mock_add_request.assert_called_once_with(expected_dict)

    @mock.patch.object(ovn_driver.OvnProviderDriver,
                       '_is_health_check_supported')
    def test_health_monitor_create_not_supported(self, ihcs):
        ihcs.return_value = False
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.health_monitor_create,
                          self.ref_health_monitor)

    def test_health_monitor_create_failure(self):
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.health_monitor_create,
                          self.fail_health_monitor)

    def test_health_monitor_create_failure_unset_type(self):
        self.fail_health_monitor.type = data_models.UnsetType()
        self.assertRaises(exceptions.UnsupportedOptionError,
                          self.driver.health_monitor_create,
                          self.fail_health_monitor)

    def test_health_monitor_create_unset_admin_state_up(self):
        self.ref_health_monitor.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_health_monitor.healthmonitor_id,
                'pool_id': self.ref_health_monitor.pool_id,
                'type': self.ref_health_monitor.type,
                'interval': self.ref_health_monitor.delay,
                'timeout': self.ref_health_monitor.timeout,
                'failure_count': self.ref_health_monitor.max_retries_down,
                'success_count': self.ref_health_monitor.max_retries,
                'admin_state_up': True}
        expected_dict = {'type': ovn_const.REQ_TYPE_HM_CREATE,
                         'info': info}
        self.driver.health_monitor_create(self.ref_health_monitor)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_health_monitor_update(self):
        info = {'id': self.ref_update_health_monitor.healthmonitor_id,
                'pool_id': self.ref_health_monitor.pool_id,
                'interval': self.ref_update_health_monitor.delay,
                'timeout': self.ref_update_health_monitor.timeout,
                'failure_count':
                    self.ref_update_health_monitor.max_retries_down,
                'success_count':
                    self.ref_update_health_monitor.max_retries,
                'admin_state_up':
                    self.ref_update_health_monitor.admin_state_up}
        expected_dict = {'type': ovn_const.REQ_TYPE_HM_UPDATE,
                         'info': info}
        self.driver.health_monitor_update(self.ref_health_monitor,
                                          self.ref_update_health_monitor)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_health_monitor_update_unset_admin_state_up(self):
        self.ref_update_health_monitor.admin_state_up = data_models.UnsetType()
        info = {'id': self.ref_update_health_monitor.healthmonitor_id,
                'pool_id': self.ref_health_monitor.pool_id,
                'interval': self.ref_update_health_monitor.delay,
                'timeout': self.ref_update_health_monitor.timeout,
                'failure_count':
                    self.ref_update_health_monitor.max_retries_down,
                'success_count':
                    self.ref_update_health_monitor.max_retries,
                'admin_state_up': True}
        expected_dict = {'type': ovn_const.REQ_TYPE_HM_UPDATE,
                         'info': info}
        self.driver.health_monitor_update(self.ref_health_monitor,
                                          self.ref_update_health_monitor)
        self.mock_add_request.assert_called_once_with(expected_dict)

    def test_health_monitor_delete(self):
        info = {'id': self.ref_health_monitor.healthmonitor_id,
                'pool_id': self.ref_health_monitor.pool_id}
        expected_dict = {'type': ovn_const.REQ_TYPE_HM_DELETE,
                         'info': info}
        self.driver.health_monitor_delete(self.ref_health_monitor)
        self.mock_add_request.assert_called_once_with(expected_dict)

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_create.assert_called_once_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
        )
        mock_listener_create.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
        )
        mock_pool_create.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]),
        )
        mock_member_create.assert_called_once_with(
            self.driver._get_member_request_info(
                self.ref_lb_fully_populated.pools[0].members[0]),
        )

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_without_listeners_or_pools(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.ref_lb_fully_populated.listeners = data_models.UnsetType()
        self.ref_lb_fully_populated.pools = data_models.UnsetType()
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_create.assert_called_once_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
        )
        mock_listener_create.assert_not_called()
        mock_pool_create.assert_not_called()

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_member_without_subnet(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.ref_lb_fully_populated.listeners = []
        self.ref_lb_fully_populated.pools[0].members[0].subnet_id = None
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_member_create.assert_called_once_with(
            self.driver._get_member_request_info(
                self.ref_lb_fully_populated.pools[0].members[0]),
        )

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'hm_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_hm_found(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_hm_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        with mock.patch.object(
                ovn_helper.OvnProviderHelper, '_find_ovn_lb_from_hm_id') \
                as mock_find_ovn_lb_from_hm_id:
            mock_find_ovn_lb_from_hm_id.return_value = (mock.ANY, self.ovn_lb)
            self.ref_pool.healthmonitor = self.ref_health_monitor
            self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
            mock_hm_create.assert_not_called()

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'hm_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_hm_lbhc_not_found(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_hm_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        with mock.patch.object(
                ovn_helper.OvnProviderHelper, '_find_ovn_lb_from_hm_id') \
                as mock_find_ovn_lb_from_hm_id:
            mock_find_ovn_lb_from_hm_id.return_value = (None, None)
            self.ref_pool.healthmonitor = self.ref_health_monitor
            self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
            mock_hm_create.assert_called_once_with(
                self.driver._get_healthmonitor_request_info(
                    self.ref_lb_fully_populated.pools[0].healthmonitor),
            )

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_without_pools(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.ref_lb_fully_populated.pools = data_models.Unset
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_create.assert_called_once_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
        )
        mock_listener_create.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
        )
        mock_pool_create.assert_not_called()

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_without_members(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.ref_lb_fully_populated.pools[0].members = []
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_create.assert_called_once_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
        )
        mock_listener_create.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
        )
        mock_pool_create.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]),
        )
        mock_member_create.assert_not_called()

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'hm_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_with_hm(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_hm_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.ref_lb_fully_populated.pools[0].members = []
        self.ref_pool.healthmonitor = self.ref_health_monitor
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_create.assert_called_once_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
        )
        mock_listener_create.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
        )
        mock_pool_create.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]),
        )
        mock_member_create.assert_not_called()
        mock_hm_create.assert_called_once_with(
            self.driver._get_healthmonitor_request_info(
                self.ref_lb_fully_populated.pools[0].healthmonitor),
        )

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_no_listeners(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.ref_lb_fully_populated.listeners = []
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_create.assert_called_once_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
        )
        mock_listener_create.assert_not_called()
        mock_pool_create.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]),
        )
        mock_member_create.assert_called_once_with(
            self.driver._get_member_request_info(
                self.ref_lb_fully_populated.pools[0].members[0]),
        )

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_create')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_create')
    def test_ensure_loadbalancer_lb_not_found_no_pools(
            self, mock_lb_create, mock_listener_create, mock_pool_create,
            mock_member_create, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.side_effect = [
            idlutils.RowNotFound]
        self.ref_lb_fully_populated.pools = []
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_create.assert_called_once_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
        )
        mock_listener_create.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
        )
        mock_pool_create.assert_not_called()
        mock_member_create.assert_not_called()

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'hm_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    def test_ensure_loadbalancer_lb_found(
            self, mock_lb_sync, mock_listener_sync, mock_pool_sync,
            mock_member_sync, mock_hm_sync, mock_get_status,
            mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        self.ref_pool.healthmonitor = self.ref_health_monitor
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_sync.assert_called_with(
            self.driver._get_loadbalancer_request_info(
                self.ref_lb_fully_populated),
            self.ovn_lb
        )
        mock_listener_sync.assert_called_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
            self.ovn_lb
        )
        mock_listener_sync.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
            self.ovn_lb
        )
        mock_pool_sync.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]),
            self.ovn_lb
        )
        mock_member_sync.assert_called_once_with(
            self.driver._get_member_request_info(
                self.ref_lb_fully_populated.pools[0].members[0]),
            mock.ANY,
            f"pool_{self.ref_lb_fully_populated.pools[0].pool_id}"
        )
        mock_hm_sync.assert_called_once_with(
            self.driver._get_healthmonitor_request_info(
                self.ref_lb_fully_populated.pools[0].healthmonitor),
            mock.ANY,
            f"pool_{self.ref_lb_fully_populated.pools[0].pool_id}"
        )
        mock_get_status.assert_called_with(self.ovn_lb)

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    def test_ensure_loadbalancer_lb_found_no_listeners_no_pools(
            self, mock_lb_sync, mock_listener_sync, mock_pool_sync,
            mock_get_status, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        self.ref_lb_fully_populated.listeners = data_models.Unset
        self.ref_lb_fully_populated.pools = data_models.Unset
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_sync.assert_called_with(
            self.driver._get_loadbalancer_request_info(self.ref_lb0),
            self.ovn_lb
        )
        mock_listener_sync.assert_not_called()
        mock_pool_sync.assert_not_called()
        mock_get_status.assert_called_with(self.ovn_lb)

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    def test_ensure_loadbalancer_lb_found_no_pools(
            self, mock_lb_sync, mock_listener_sync, mock_pool_sync,
            mock_get_status, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        self.ref_lb_fully_populated.pools = data_models.Unset
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_listener_sync.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
            self.ovn_lb
        )
        mock_lb_sync.assert_called_with(
            self.driver._get_loadbalancer_request_info(self.ref_lb0),
            self.ovn_lb
        )
        mock_pool_sync.assert_not_called()
        mock_get_status.assert_called_with(self.ovn_lb)

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'hm_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    @mock.patch('ovn_octavia_provider.common.clients.get_neutron_client')
    def test_ensure_loadbalancer_lb_found_no_hm(
            self, net_cli, mock_lb_sync, mock_listener_sync,
            mock_pool_sync, mock_member_sync, mock_hm_sync,
            mock_get_status, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        self.ref_pool_with_hm.healthmonitor = None
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)

        mock_lb_sync.assert_called_with(
            self.driver._get_loadbalancer_request_info(self.ref_lb0),
            self.ovn_lb
        )
        mock_listener_sync.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
            self.ovn_lb
        )
        mock_pool_sync.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]),
            self.ovn_lb
        )
        mock_hm_sync.assert_not_called()

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    def test_ensure_loadbalancer_lb_found_no_members(
            self, mock_lb_sync, mock_listener_sync, mock_pool_sync,
            mock_member_sync, mock_get_status, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        self.ovn_lb.external_ids.pop('pool_%s' % self.pool_id)
        self.ref_lb_fully_populated.pools[0].members = data_models.Unset
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)

        mock_lb_sync.assert_called_with(
            self.driver._get_loadbalancer_request_info(self.ref_lb0),
            self.ovn_lb
        )
        mock_listener_sync.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
            self.ovn_lb
        )
        mock_pool_sync.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]), self.ovn_lb)
        mock_member_sync.assert_not_called()

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    def test_ensure_loadbalancer_lb_found_member_without_subnet(
            self, mock_lb_sync, mock_listener_sync, mock_pool_sync,
            mock_member_sync, mock_get_status, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        self.ref_lb_fully_populated.listeners = []
        self.ref_lb_fully_populated.pools[0].members[0].subnet_id = None
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_member_sync.assert_called_once_with(
            self.driver._get_member_request_info(
                self.ref_lb_fully_populated.pools[0].members[0]),
            mock.ANY,
            f"pool_{self.ref_lb_fully_populated.pools[0].pool_id}"
        )

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    def test_ensure_loadbalancer_lb_found_no_members_to_delete(
            self, mock_lb_sync, mock_listener_sync, mock_pool_sync,
            mock_member_sync, mock_get_status, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        self.ovn_lb.external_ids.pop('pool_%s' % self.pool_id)
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_sync.assert_called_with(
            self.driver._get_loadbalancer_request_info(self.ref_lb0),
            self.ovn_lb
        )
        mock_listener_sync.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
            self.ovn_lb
        )
        mock_pool_sync.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]), self.ovn_lb)
        mock_member_sync.assert_called_once_with(
            self.driver._get_member_request_info(
                self.ref_lb_fully_populated.pools[0].members[0]),
            mock.ANY,
            f"pool_{self.ref_lb_fully_populated.pools[0].pool_id}"
        )

    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_update_status_to_octavia')
    @mock.patch.object(ovn_helper.OvnProviderHelper,
                       '_get_current_operating_statuses')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'handle_member_dvr')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_delete')
    @mock.patch.object(ovn_helper.OvnProviderHelper, '_get_members_in_ovn_lb')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'member_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'pool_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'listener_sync')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'lb_sync')
    def test_ensure_loadbalancer_lb_found_one_member_to_delete(
            self, mock_lb_sync, mock_listener_sync, mock_pool_sync,
            mock_member_sync, mock_get_members_in_ovn_lb, mock_member_delete,
            mock_handle_member_dvr, mock_get_status, mock_update_status):
        self.mock_find_ovn_lbs_with_retry.return_value = [
            self.ovn_lb]
        mock_get_members_in_ovn_lb.return_value = [
            ['ip', 'port', 'subnet', 'foo']
        ]
        self.ovn_lb.external_ids.pop('pool_%s' % self.pool_id)
        self.driver._ensure_loadbalancer(self.ref_lb_fully_populated)
        mock_lb_sync.assert_called_with(
            self.driver._get_loadbalancer_request_info(self.ref_lb0),
            self.ovn_lb
        )
        mock_listener_sync.assert_called_once_with(
            self.driver._get_listener_request_info(
                self.ref_lb_fully_populated.listeners[0]),
            self.ovn_lb
        )
        mock_pool_sync.assert_called_once_with(
            self.driver._get_pool_request_info(
                self.ref_lb_fully_populated.pools[0]), self.ovn_lb)
        mock_member_sync.assert_called_once_with(
            self.driver._get_member_request_info(
                self.ref_lb_fully_populated.pools[0].members[0]),
            mock.ANY,
            f"pool_{self.ref_lb_fully_populated.pools[0].pool_id}"
        )
        mock_member_delete.assert_called_once_with({
            'id': 'foo',
            'subnet_id': 'subnet'}
        )

    @mock.patch.object(ovn_helper.OvnProviderHelper, 'get_octavia_lbs')
    @mock.patch.object(clients, 'get_octavia_client')
    def test_do_sync_no_loadbalancers(self, mock_get_octavia_client,
                                      mock_get_octavia_lbs, ):
        mock_get_octavia_lbs.return_value = []
        lb_filters = {}
        with mock.patch.object(clients, 'get_neutron_client',
                               side_effect=RuntimeError), \
                mock.patch.object(self.driver,
                                  '_ensure_loadbalancer') as mock_ensure_lb:
            self.driver.do_sync(**lb_filters)
            mock_ensure_lb.assert_not_called()

    @mock.patch.object(data_models.HealthMonitor, 'from_dict')
    @mock.patch.object(data_models.Member, 'from_dict')
    @mock.patch.object(data_models.Listener, 'from_dict')
    @mock.patch.object(data_models.Pool, 'from_dict')
    @mock.patch.object(o_driver_lib.DriverLibrary, 'get_loadbalancer')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'get_octavia_lbs')
    @mock.patch.object(clients, 'get_octavia_client')
    def test_do_sync_with_loadbalancers(self,
                                        mock_get_octavia_client,
                                        mock_get_octavia_lbs,
                                        mock_get_loadbalancer,
                                        mock_pool_from_dict,
                                        mock_listener_from_dict,
                                        mock_member_from_dict,
                                        mock_hm_from_dict):
        lb = mock.MagicMock(id=self.ref_lb_fully_sync_populated.name)
        mock_get_octavia_lbs.return_value = [lb]
        mock_get_loadbalancer.return_value = self.ref_lb_fully_sync_populated
        mock_pool_from_dict.return_value = self.ref_pool_with_hm
        mock_listener_from_dict.return_value = self.ref_listener
        mock_member_from_dict.return_value = self.ref_member
        mock_hm_from_dict.return_value = self.ref_health_monitor
        lb_filters = {}
        with mock.patch.object(self.driver, '_ensure_loadbalancer') \
                as mock_ensure_lb:
            self.driver.do_sync(**lb_filters)
            mock_ensure_lb.assert_any_call(
                self.ref_lb_fully_sync_populated)

    @mock.patch.object(data_models.HealthMonitor, 'from_dict')
    @mock.patch.object(data_models.Member, 'from_dict')
    @mock.patch.object(data_models.Listener, 'from_dict')
    @mock.patch.object(data_models.Pool, 'from_dict')
    @mock.patch.object(o_driver_lib.DriverLibrary, 'get_loadbalancer')
    @mock.patch.object(ovn_helper.OvnProviderHelper, 'get_octavia_lbs')
    @mock.patch.object(clients, 'get_octavia_client')
    def test_do_sync_with_loadbalancer_no_members_no_hm(
            self, mock_get_octavia_client, mock_get_octavia_lbs,
            mock_get_loadbalancer, mock_pool_from_dict,
            mock_listener_from_dict, mock_member_from_dict, mock_hm_from_dict):
        self.ref_pool_with_hm.members = []
        self.ref_lb_fully_sync_populated.pools[0].healthmonitor = None
        lb = mock.MagicMock(id=self.ref_lb_fully_sync_populated.name)
        mock_get_octavia_lbs.return_value = [lb]
        mock_get_loadbalancer.return_value = self.ref_lb_fully_sync_populated
        mock_pool_from_dict.return_value = self.ref_pool_with_hm
        mock_listener_from_dict.return_value = self.ref_listener
        lb_filters = {}
        with mock.patch.object(self.driver, '_ensure_loadbalancer') \
                as mock_ensure_lb:
            self.driver.do_sync(**lb_filters)
            mock_ensure_lb.assert_any_call(
                self.ref_lb_fully_sync_populated)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging

from ovn_octavia_provider.i18n import _

LOG = logging.getLogger(__name__)


ovn_opts = [
    cfg.ListOpt('ovn_nb_connection',
                default=['tcp:127.0.0.1:6641'],
                item_type=types.String(regex=r'^(tcp|ssl|unix):.+'),
                help=_('The connection string for the OVN_Northbound OVSDB.\n'
                       'Use tcp:IP:PORT for TCP connection.\n'
                       'Use ssl:IP:PORT for SSL connection. The '
                       'ovn_nb_private_key, ovn_nb_certificate and '
                       'ovn_nb_ca_cert are mandatory.\n'
                       'Use unix:FILE for unix domain socket connection.')),
    cfg.StrOpt('ovn_nb_private_key',
               default='',
               help=_('The PEM file with private key for SSL connection to '
                      'OVN-NB-DB')),
    cfg.StrOpt('ovn_nb_certificate',
               default='',
               help=_('The PEM file with certificate that certifies the '
                      'private key specified in ovn_nb_private_key')),
    cfg.StrOpt('ovn_nb_ca_cert',
               default='',
               help=_('The PEM file with CA certificate that OVN should use to'
                      ' verify certificates presented to it by SSL peers')),
    cfg.ListOpt('ovn_sb_connection',
                default=['tcp:127.0.0.1:6642'],
                item_type=types.String(regex=r'^(tcp|ssl|unix):.+'),
                help=_('The connection string for the OVN_Southbound OVSDB.\n'
                       'Use tcp:IP:PORT for TCP connection.\n'
                       'Use ssl:IP:PORT for SSL connection. The '
                       'ovn_sb_private_key, ovn_sb_certificate and '
                       'ovn_sb_ca_cert are mandatory.\n'
                       'Use unix:FILE for unix domain socket connection.')),
    cfg.StrOpt('ovn_sb_private_key',
               default='',
               help=_('The PEM file with private key for SSL connection to '
                      'OVN-SB-DB')),
    cfg.StrOpt('ovn_sb_certificate',
               default='',
               help=_('The PEM file with certificate that certifies the '
                      'private key specified in ovn_sb_private_key')),
    cfg.StrOpt('ovn_sb_ca_cert',
               default='',
               help=_('The PEM file with CA certificate that OVN should use to'
                      ' verify certificates presented to it by SSL peers')),
    cfg.IntOpt('ovsdb_connection_timeout',
               default=180,
               help=_('Timeout in seconds for the OVSDB '
                      'connection transaction')),
    cfg.IntOpt('ovsdb_retry_max_interval',
               default=180,
               help=_('Max interval in seconds between '
                      'each retry to get the OVN NB and SB IDLs')),
    cfg.IntOpt('ovsdb_probe_interval',
               min=0,
               default=60000,
               help=_('The probe interval in for the OVSDB session in '
                      'milliseconds. If this is zero, it disables the '
                      'connection keepalive feature. If non-zero the value '
                      'will be forced to at least 1000 milliseconds. Defaults '
                      'to 60 seconds.')),
]

neutron_opts = [
    cfg.StrOpt('endpoint', help=_('A new endpoint to override the endpoint '
                                  'in the keystone catalog.'),
               deprecated_for_removal=True,
               deprecated_reason=_('The endpoint_override option defined by '
                                   'keystoneauth1 is the new name for this '
                                   'option.'),
               deprecated_since='Antelope'),
    cfg.StrOpt('endpoint_type', help=_('Endpoint interface in identity '
                                       'service to use'),
               deprecated_for_removal=True,
               deprecated_reason=_('This option was replaced by the '
                                   'valid_interfaces option defined by '
                                   'keystoneauth.'),
               deprecated_since='Antelope'),
    cfg.StrOpt('ca_certificates_file',
               help=_('CA certificates file path'),
               deprecated_for_removal=True,
               deprecated_reason=_('The cafile option defined by '
                                   'keystoneauth1 is the new name for this '
                                   'option.'),
               deprecated_since='Antelope'),
]


def handle_neutron_deprecations():
    # Apply neutron deprecated options to their new setting if needed

    # Basicaly: if the value of the deprecated option is not the default:
    # * convert it to a valid "new" value if needed
    # * set it as the default for the new option
    # Thus [neutron].<new_option> has an higher precedence than
    # [neutron].<deprecated_option>
    loc = cfg.CONF.get_location('endpoint', 'neutron')
    if loc and loc.location != cfg.Locations.opt_default:
        cfg.CONF.set_default('endpoint_override', cfg.CONF.neutron.endpoint,
                             'neutron')

    loc = cfg.CONF.get_location('endpoint_type', 'neutron')
    if loc and loc.location != cfg.Locations.opt_default:
        endpoint_type = cfg.CONF.neutron.endpoint_type.replace('URL', '')
        cfg.CONF.set_default('valid_interfaces', [endpoint_type],
                             'neutron')

    loc = cfg.CONF.get_location('ca_certificates_file', 'neutron')
    if loc and loc.location != cfg.Locations.opt_default:
        cfg.CONF.set_default('cafile', cfg.CONF.neutron.ca_certificates_file,
                             'neutron')


def register_opts():
    # NOTE (froyo): just to not try to re-register options already done
    # by Neutron, specially in test scope, that will get a DuplicateOptError
    missing_ovn_opts = ovn_opts
    try:
        neutron_registered_opts = [opt for opt in cfg.CONF.ovn]
        missing_ovn_opts = [opt for opt in ovn_opts
                            if opt.name not in neutron_registered_opts]
    except cfg.NoSuchOptError:
        LOG.info('Not found any opts under group ovn registered by Neutron')

    # Do the same for neutron options that have been already registered by
    # Octavia
    missing_neutron_opts = neutron_opts
    try:
        neutron_registered_opts = [opt for opt in cfg.CONF.neutron]
        missing_neutron_opts = [opt for opt in neutron_opts
                                if opt.name not in neutron_registered_opts]
    except cfg.NoSuchOptError:
        LOG.info('Not found any opts under group neutron')

    cfg.CONF.register_opts(missing_ovn_opts, group='ovn')
    cfg.CONF.register_opts(missing_neutron_opts, group='neutron')
    ks_loading.register_auth_conf_options(cfg.CONF, 'service_auth')
    ks_loading.register_session_conf_options(cfg.CONF, 'service_auth')
    ks_loading.register_adapter_conf_options(cfg.CONF, 'service_auth',
                                             include_deprecated=False)

    ks_loading.register_auth_conf_options(cfg.CONF, 'neutron')
    ks_loading.register_session_conf_options(cfg.CONF, 'neutron')
    ks_loading.register_adapter_conf_options(cfg.CONF, 'neutron',
                                             include_deprecated=False)

    # Override default auth_type for plugins with the default from service_auth
    auth_type = cfg.CONF.service_auth.auth_type
    cfg.CONF.set_default('auth_type', auth_type, 'neutron')

    handle_neutron_deprecations()


def list_opts():
    return [
        ('ovn', ovn_opts),
        ('neutron', neutron_opts),
    ]


def get_ovn_nb_connection():
    return ','.join(cfg.CONF.ovn.ovn_nb_connection)


def get_ovn_nb_private_key():
    return cfg.CONF.ovn.ovn_nb_private_key


def get_ovn_nb_certificate():
    return cfg.CONF.ovn.ovn_nb_certificate


def get_ovn_nb_ca_cert():
    return cfg.CONF.ovn.ovn_nb_ca_cert


def get_ovn_sb_connection():
    return ','.join(cfg.CONF.ovn.ovn_sb_connection)


def get_ovn_sb_private_key():
    return cfg.CONF.ovn.ovn_sb_private_key


def get_ovn_sb_certificate():
    return cfg.CONF.ovn.ovn_sb_certificate


def get_ovn_sb_ca_cert():
    return cfg.CONF.ovn.ovn_sb_ca_cert


def get_ovn_ovsdb_timeout():
    return cfg.CONF.ovn.ovsdb_connection_timeout


def get_ovn_ovsdb_retry_max_interval():
    return cfg.CONF.ovn.ovsdb_retry_max_interval


def get_ovn_ovsdb_probe_interval():
    return cfg.CONF.ovn.ovsdb_probe_interval



#    Copyright 2023 Red Hat, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import threading

from futurist import periodics
import netaddr
from neutron_lib import constants as n_const
from oslo_config import cfg
from oslo_log import log as logging
from ovsdbapp.backend.ovs_idl import connection

from ovn_octavia_provider.common import clients
from ovn_octavia_provider.common import config as ovn_conf
from ovn_octavia_provider.common import constants as ovn_const
from ovn_octavia_provider.ovsdb import impl_idl_ovn

CONF = cfg.CONF  # Gets Octavia Conf as it runs under o-api domain

LOG = logging.getLogger(__name__)


class MaintenanceThread(object):

    def __init__(self):
        self._callables = []
        self._thread = None
        self._worker = None

    def add_periodics(self, obj):
        for name, member in inspect.getmembers(obj):
            if periodics.is_periodic(member):
                LOG.info('Periodic task found: %(owner)s.%(member)s',
                         {'owner': obj.__class__.__name__, 'member': name})
                self._callables.append((member, (), {}))

    def start(self):
        if self._thread is None:
            self._worker = periodics.PeriodicWorker(self._callables)
            self._thread = threading.Thread(target=self._worker.start)
            self._thread.daemon = True
            self._thread.start()

    def stop(self):
        self._worker.stop()
        self._worker.wait()
        self._thread.join()
        self._worker = self._thread = None


class DBInconsistenciesPeriodics(object):

    def __init__(self):
        self.ovn_nbdb = impl_idl_ovn.OvnNbIdlForLb()
        c = connection.Connection(self.ovn_nbdb,
                                  ovn_conf.get_ovn_ovsdb_timeout())
        self.ovn_nbdb_api = impl_idl_ovn.OvsdbNbOvnIdl(c)

    @periodics.periodic(spacing=600, run_immediately=True)
    def change_device_owner_lb_hm_ports(self):
        """Change the device_owner for the OVN LB HM port existing.

        The OVN LB HM port used for send the health checks to the backend
        members has a new device_owner, it will use the value
        onv-lb-hm:distributed in order to keep the behaviour on Neutron as a
        LOCALPORT. Also this change will add device-id as ovn-lb-hm:{subnet}
        to get more robust.
        """
        LOG.debug('Maintenance task: checking device_owner for OVN LB HM '
                  'ports.')
        neutron_client = clients.get_neutron_client()
        ovn_lb_hm_ports = neutron_client.ports(
            device_owner=n_const.DEVICE_OWNER_DISTRIBUTED)

        check_neutron_support_new_device_owner = True
        for port in ovn_lb_hm_ports:
            if port.name.startswith('ovn-lb-hm'):
                LOG.debug('Maintenance task: updating device_owner and '
                          'adding device_id for port id %s', port.id)
                neutron_client.update_port(
                    port.id, device_owner=ovn_const.OVN_LB_HM_PORT_DISTRIBUTED,
                    device_id=port.name)

                # NOTE(froyo): Check that the port is now of type LOCALPORT in
                # the OVN NB DB or perform a rollback in other cases. Such
                # cases could indicate that Neutron is in the process of being
                # updated or that the user has forgotten to update Neutron to a
                # version that supports this change
                if check_neutron_support_new_device_owner:
                    port_ovn = self.ovn_nbdb_api.db_find_rows(
                        "Logical_Switch_Port", ("name", "=", port.id)).execute(
                        check_error=True)
                    if len(port_ovn) and port_ovn[0].type != 'localport':
                        LOG.debug('Maintenance task: port %s updated but '
                                  'looks like Neutron does not support this '
                                  'new device_owner, or maybe is updating '
                                  'version, so restoring to old values and '
                                  'waiting another iteration of this task',
                                  port.id)
                        neutron_client.update_port(
                            port.id,
                            device_owner=n_const.DEVICE_OWNER_DISTRIBUTED,
                            device_id='')
                        # Break the loop as do not make sense change the rest
                        break
                    check_neutron_support_new_device_owner = False
        else:
            # NOTE(froyo): No ports found to update, or all of them done.
            LOG.debug('Maintenance task: no more ports left, stopping the '
                      'periodic task.')
            raise periodics.NeverAgain()
        LOG.debug('Maintenance task: device_owner and device_id checked for '
                  'OVN LB HM ports.')

    # TODO(froyo): Remove this in the Caracal+4 cycle
    @periodics.periodic(spacing=600, run_immediately=True)
    def format_ip_port_mappings_ipv6(self):
        """Give correct format to `ip_port_mappings` for IPv6 backend members.

        The `ip_port_mappings` field for OVN LBs should be a dictionary with
        keys following the format:
        `${MEMBER_IP}=${LSP_NAME_MEMBER}:${HEALTH_SRC_IP}`. However, when
        `MEMBER_IP` and `HEALTH_SRC_IP` are IPv6 addresses, they should be
        enclosed in `[]`.
        """
        LOG.debug('Maintenance task: Ensure correct formatting of '
                  'ip_port_mappings for IPv6 backend members.')
        ovn_lbs = self.ovn_nbdb_api.db_find_rows(
            'Load_Balancer', ('ip_port_mappings', '!=', {})).execute()
        for lb in ovn_lbs:
            mappings = {}
            for k, v in lb.ip_port_mappings.items():
                try:
                    # If first element is IPv4 (mixing IPv4 and IPv6 not
                    # allowed) or get AddrFormatError (IPv6 already fixed) we
                    # can jump to next item
                    if netaddr.IPNetwork(k).version == n_const.IP_VERSION_4:
                        break
                except netaddr.AddrFormatError:
                    break
                port_uuid, src_ip = v.split(':', 1)
                mappings[f'[{k}]'] = f'{port_uuid}:[{src_ip}]'
            if mappings:
                self.ovn_nbdb_api.db_clear(
                    'Load_Balancer', lb.uuid, 'ip_port_mappings'
                ).execute(check_error=True)
                self.ovn_nbdb_api.db_set(
                    'Load_Balancer', lb.uuid, ('ip_port_mappings', mappings)
                ).execute(check_error=True)

        LOG.debug('Maintenance task: no more ip_port_mappings to format, '
                  'stopping the periodic task.')
        raise periodics.NeverAgain()



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Guidelines for writing new hacking checks

 - Use only for OVN Octavia provider specific tests. OpenStack
   general tests should be submitted to the common 'hacking' module.
 - Pick numbers in the range N3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the N3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to
   ovn_octavia_provider/tests/unit/hacking/test_checks.py

"""

import re

from hacking import core


unittest_imports_dot = re.compile(r"\bimport[\s]+unittest\b")
unittest_imports_from = re.compile(r"\bfrom[\s]+unittest\b")
filter_match = re.compile(r".*filter\(lambda ")

tests_imports_dot = re.compile(r"\bimport[\s]+ovn_octavia_provider.tests\b")
tests_imports_from1 = re.compile(r"\bfrom[\s]+ovn_octavia_provider.tests\b")
tests_imports_from2 = re.compile(
    r"\bfrom[\s]+ovn_octavia_provider[\s]+import[\s]+tests\b")
no_line_continuation_backslash_re = re.compile(r'.*(\\)\n')

import_mock = re.compile(r"\bimport[\s]+mock\b")  # noqa: H216
import_from_mock = re.compile(r"\bfrom[\s]+mock[\s]+import\b")


@core.flake8ext
def check_assert_called_once_with(logical_line, filename):
    """Try to detect unintended calls of nonexistent mock methods like:

                 assert_called_once
                 assertCalledOnceWith
                 assert_has_called
                 called_once_with

    N322
    """

    if 'ovn_octavia_provider/tests/' in filename:
        if '.assert_called_once_with(' in logical_line:
            return
        uncased_line = logical_line.lower().replace('_', '')

        check_calls = ['.assertcalledonce', '.calledoncewith']
        if any(x for x in check_calls if x in uncased_line):
            msg = ("N322: Possible use of no-op mock method. "
                   "please use assert_called_once_with.")
            yield (0, msg)

        if '.asserthascalled' in uncased_line:
            msg = ("N322: Possible use of no-op mock method. "
                   "please use assert_has_calls.")
            yield (0, msg)


@core.flake8ext
def check_asserttruefalse(logical_line, filename):
    """N328 - Don't use assertEqual(True/False, observed)."""

    if 'ovn_octavia_provider/tests/' in filename:
        if re.search(r"assertEqual\(\s*True,[^,]*(,[^,]*)?", logical_line):
            msg = ("N328: Use assertTrue(observed) instead of "
                   "assertEqual(True, observed)")
            yield (0, msg)
        if re.search(r"assertEqual\([^,]*,\s*True(,[^,]*)?", logical_line):
            msg = ("N328: Use assertTrue(observed) instead of "
                   "assertEqual(True, observed)")
            yield (0, msg)
        if re.search(r"assertEqual\(\s*False,[^,]*(,[^,]*)?", logical_line):
            msg = ("N328: Use assertFalse(observed) instead of "
                   "assertEqual(False, observed)")
            yield (0, msg)
        if re.search(r"assertEqual\([^,]*,\s*False(,[^,]*)?", logical_line):
            msg = ("N328: Use assertFalse(observed) instead of "
                   "assertEqual(False, observed)")
            yield (0, msg)


@core.flake8ext
def check_assertempty(logical_line, filename):
    """Enforce using assertEqual parameter ordering in case of empty objects.

    N330
    """

    if 'ovn_octavia_provider/tests/' in filename:
        msg = ("N330: Use assertEqual(*empty*, observed) instead of "
               "assertEqual(observed, *empty*). *empty* contains "
               "{}, [], (), set(), '', \"\"")
        empties = r"(\[\s*\]|\{\s*\}|\(\s*\)|set\(\s*\)|'\s*'|\"\s*\")"
        reg = fr"assertEqual\(([^,]*,\s*)+?{empties}\)\s*$"
        if re.search(reg, logical_line):
            yield (0, msg)


@core.flake8ext
def check_assertisinstance(logical_line, filename):
    """N331 - Enforce using assertIsInstance."""

    if 'ovn_octavia_provider/tests/' in filename:
        if re.search(r"assertTrue\(\s*isinstance\(\s*[^,]*,\s*[^,]*\)\)",
                     logical_line):
            msg = ("N331: Use assertIsInstance(observed, type) instead "
                   "of assertTrue(isinstance(observed, type))")
            yield (0, msg)


@core.flake8ext
def check_assertequal_for_httpcode(logical_line, filename):
    """N332 - Enforce correct oredering for httpcode in assertEqual."""

    msg = ("N332: Use assertEqual(expected_http_code, observed_http_code) "
           "instead of assertEqual(observed_http_code, expected_http_code)")
    if 'ovn_octavia_provider/tests/' in filename:
        if re.search(r"assertEqual\(\s*[^,]*,[^,]*HTTP[^\.]*\.code\s*\)",
                     logical_line):
            yield (0, msg)


@core.flake8ext
def check_no_imports_from_tests(logical_line, filename):
    """N343 - Production code must not import from ovn_octavia_provider.tests.*

    """

    msg = ("N343 Production code must not import from "
           "ovn_octavia_provider.tests.*")

    if 'ovn_octavia_provider/tests/' in filename:
        return

    for regex in tests_imports_dot, tests_imports_from1, tests_imports_from2:
        if re.match(regex, logical_line):
            yield (0, msg)


@core.flake8ext
def check_python3_no_filter(logical_line):
    """N344 - Use list comprehension instead of filter(lambda)."""

    msg = ("N344: Use list comprehension instead of "
           "filter(lambda obj: test(obj), data) on python3.")

    if filter_match.match(logical_line):
        yield (0, msg)


@core.flake8ext
def check_no_import_mock(logical_line, filename, noqa):
    """N347 - Test code must not import mock library."""
    msg = ("N347: Test code must not import mock library")

    if noqa:
        return

    if 'ovn_octavia_provider/tests/' not in filename:
        return

    for regex in import_mock, import_from_mock:
        if re.match(regex, logical_line):
            yield (0, msg)


@core.flake8ext
def check_assertcountequal(logical_line, filename):
    """N348 - Enforce using assertCountEqual."""

    msg = ("N348: Use assertCountEqual(expected, observed) "
           "instead of assertItemsEqual(observed, expected)")

    if 'ovn_octavia_provider/tests/' in filename:
        if re.search(r"assertItemsEqual\([^,]*,\s*(,[^,]*)?", logical_line):
            yield (0, msg)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import os
from unittest import mock

from neutron.tests import base
from ovs.db import idl as ovs_idl
from ovsdbapp.backend import ovs_idl as real_ovs_idl
from ovsdbapp.backend.ovs_idl import idlutils

from ovn_octavia_provider.common import config as ovn_config
from ovn_octavia_provider.ovsdb import impl_idl_ovn

basedir = os.path.dirname(os.path.abspath(__file__))
schema_files = {
    'OVN_Northbound': os.path.join(basedir,
                                   '..', 'schemas', 'ovn-nb.ovsschema'),
    'OVN_Southbound': os.path.join(basedir,
                                   '..', 'schemas', 'ovn-sb.ovsschema')}


class TestOvnNbIdlForLb(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        ovn_config.register_opts()
        self.mock_gsh = mock.patch.object(
            idlutils, 'get_schema_helper',
            side_effect=lambda x, y: ovs_idl.SchemaHelper(
                location=schema_files['OVN_Northbound'])).start()
        self.idl = impl_idl_ovn.OvnNbIdlForLb()

    def test__get_ovsdb_helper(self):
        self.mock_gsh.reset_mock()
        self.idl._get_ovsdb_helper('foo')
        self.mock_gsh.assert_called_once_with('foo', 'OVN_Northbound')

    def test_setlock(self):
        with mock.patch.object(impl_idl_ovn.OvnNbIdlForLb,
                               'set_lock') as set_lock:
            self.idl = impl_idl_ovn.OvnNbIdlForLb(event_lock_name='foo')
        set_lock.assert_called_once_with('foo')


class TestOvnSbIdlForLb(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        ovn_config.register_opts()
        self.mock_gsh = mock.patch.object(
            idlutils, 'get_schema_helper',
            side_effect=lambda x, y: ovs_idl.SchemaHelper(
                location=schema_files['OVN_Southbound'])).start()
        self.idl = impl_idl_ovn.OvnSbIdlForLb()

    @mock.patch.object(real_ovs_idl.Backend, 'autocreate_indices', mock.Mock(),
                       create=True)
    def test_start_reuses_connection(self):
        with mock.patch('ovsdbapp.backend.ovs_idl.connection.Connection',
                        side_effect=lambda x, timeout: mock.Mock()):
            idl1 = impl_idl_ovn.OvnSbIdlForLb()
            ret1 = idl1.start()
            id1 = id(ret1.ovsdb_connection)
            idl2 = impl_idl_ovn.OvnSbIdlForLb()
            ret2 = idl2.start()
            id2 = id(ret2.ovsdb_connection)
            self.assertEqual(id1, id2)

    @mock.patch('ovsdbapp.backend.ovs_idl.connection.Connection')
    def test_stop(self, mock_conn):
        mock_conn.stop.return_value = False
        with (
            mock.patch.object(
                self.idl.notify_handler, 'shutdown')) as mock_notify, (
                mock.patch.object(self.idl, 'close')) as mock_close:
            self.idl.start()
            self.idl.stop()
        mock_notify.assert_called_once_with()
        mock_close.assert_called_once_with()

    @mock.patch('ovsdbapp.backend.ovs_idl.connection.Connection')
    def test_stop_no_connection(self, mock_conn):
        mock_conn.stop.return_value = False
        with (
            mock.patch.object(
                self.idl.notify_handler, 'shutdown')) as mock_notify, (
                mock.patch.object(self.idl, 'close')) as mock_close:
            self.idl.stop()
        mock_notify.assert_called_once_with()
        mock_close.assert_called_once_with()

    def test_setlock(self):
        with mock.patch.object(impl_idl_ovn.OvnSbIdlForLb,
                               'set_lock') as set_lock:
            self.idl = impl_idl_ovn.OvnSbIdlForLb(event_lock_name='foo')
        set_lock.assert_called_once_with('foo')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octavia_lib.common import constants


# TODO(mjozefcz): Use those variables from neutron-lib once released.
LRP_PREFIX = "lrp-"
OVN_NAME_PREFIX = "neutron-"
LB_HM_PORT_PREFIX = "ovn-lb-hm-"
LB_VIP_PORT_PREFIX = "ovn-lb-vip-"
LB_VIP_ADDIT_PORT_PREFIX = "ovn-lb-vip-additional-"
OVN_PORT_NAME_EXT_ID_KEY = 'neutron:port_name'
OVN_ROUTER_NAME_EXT_ID_KEY = 'neutron:router_name'
OVN_PORT_FIP_EXT_ID_KEY = 'neutron:port_fip'
OVN_SUBNET_EXT_ID_KEY = 'neutron:subnet_id'
OVN_SUBNET_EXT_IDS_KEY = 'neutron:subnet_ids'
OVN_NETWORK_NAME_EXT_ID_KEY = 'neutron:network_name'
OVN_PROJECT_EXT_ID_KEY = 'neutron:project_id'
OVN_SG_IDS_EXT_ID_KEY = 'neutron:security_group_ids'
OVN_DEVICE_OWNER_EXT_ID_KEY = 'neutron:device_owner'
OVN_FIP_EXT_ID_KEY = 'neutron:fip_id'
OVN_FIP_PORT_EXT_ID_KEY = 'neutron:fip_port_id'
OVN_GW_PORT_EXT_ID_KEY = 'neutron:gw_port_id'
OVN_PORT_CIDR_EXT_ID_KEY = 'neutron:cidrs'
OVN_MEMBER_STATUS_KEY = 'neutron:member_status'
OVN_ROUTER_IS_EXT_GW = 'neutron:is_ext_gw'

# TODO(froyo): Use from neutron-lib once released.
OVN_LB_HM_PORT_DISTRIBUTED = 'ovn-lb-hm:distributed'

LB_EXT_IDS_LS_REFS_KEY = 'ls_refs'
LB_EXT_IDS_LR_REF_KEY = 'lr_ref'
LB_EXT_IDS_POOL_PREFIX = 'pool_'
LB_EXT_IDS_LISTENER_PREFIX = 'listener_'
LB_EXT_IDS_MEMBER_PREFIX = 'member_'
LB_EXT_IDS_HM_KEY = 'octavia:healthmonitor'
LB_EXT_IDS_HM_POOL_KEY = 'octavia:pool_id'
LB_EXT_IDS_HM_VIP = 'octavia:vip'
LB_EXT_IDS_HMS_KEY = 'octavia:healthmonitors'

# NOTE(froyo):from additional-vips feature we will mantain the old ones for
# backward compatibility
LB_EXT_IDS_VIP_KEY = 'neutron:vip'
LB_EXT_IDS_ADDIT_VIP_KEY = 'neutron:additional_vips'
LB_EXT_IDS_VIP_PORT_ID_KEY = 'neutron:vip_port_id'
LB_EXT_IDS_ADDIT_VIP_PORT_ID_KEY = 'neutron:additional_vip_port_ids'
LB_EXT_IDS_VIP_FIP_KEY = 'neutron:vip_fip'
LB_EXT_IDS_ADDIT_VIP_FIP_KEY = 'neutron:additional_vip_fips'

PORT_FORWARDING_PLUGIN = 'port_forwarding_plugin'

# Auth sections
SERVICE_AUTH = 'service_auth'

# Request type constants
REQ_TYPE_LB_CREATE = 'lb_create'
REQ_TYPE_LB_DELETE = 'lb_delete'
REQ_TYPE_LB_UPDATE = 'lb_update'
REQ_TYPE_LISTENER_CREATE = 'listener_create'
REQ_TYPE_LISTENER_DELETE = 'listener_delete'
REQ_TYPE_LISTENER_UPDATE = 'listener_update'
REQ_TYPE_POOL_CREATE = 'pool_create'
REQ_TYPE_POOL_DELETE = 'pool_delete'
REQ_TYPE_POOL_UPDATE = 'pool_update'
REQ_TYPE_MEMBER_CREATE = 'member_create'
REQ_TYPE_MEMBER_DELETE = 'member_delete'
REQ_TYPE_MEMBER_UPDATE = 'member_update'
REQ_TYPE_LB_CREATE_LRP_ASSOC = 'lb_create_lrp_assoc'
REQ_TYPE_LB_DELETE_LRP_ASSOC = 'lb_delete_lrp_assoc'
REQ_TYPE_HANDLE_VIP_FIP = 'handle_vip_fip'
REQ_TYPE_HANDLE_MEMBER_DVR = 'handle_member_dvr'
REQ_TYPE_HM_CREATE = 'hm_create'
REQ_TYPE_HM_UPDATE = 'hm_update'
REQ_TYPE_HM_DELETE = 'hm_delete'
REQ_TYPE_HM_UPDATE_EVENT = 'hm_update_event'

REQ_TYPE_EXIT = 'exit'

# Request information constants
REQ_INFO_ACTION_ASSOCIATE = 'associate'
REQ_INFO_ACTION_SYNC = 'sync'
REQ_INFO_ACTION_DISASSOCIATE = 'disassociate'

REQ_INFO_MEMBER_ADDED = 'member_added'
REQ_INFO_MEMBER_DELETED = 'member_deleted'

# Disabled resources have a ':D' at the end
DISABLED_RESOURCE_SUFFIX = 'D'

# This driver only supports TCP, UDP and SCTP, with a single LB algorithm
OVN_NATIVE_LB_PROTOCOLS = [constants.PROTOCOL_TCP,
                           constants.PROTOCOL_UDP,
                           constants.PROTOCOL_SCTP, ]
OVN_NATIVE_LB_ALGORITHMS = [constants.LB_ALGORITHM_SOURCE_IP_PORT, ]

# This driver only supports UDP Connect and TCP health monitors
SUPPORTED_HEALTH_MONITOR_TYPES = [constants.HEALTH_MONITOR_UDP_CONNECT,
                                  constants.HEALTH_MONITOR_TCP]

# Prepended to exception log messages
EXCEPTION_MSG = "Exception occurred during %s"

# Used in functional tests
LR_REF_KEY_HEADER = 'neutron-'

# LB selection fields to represent LB algorithm
LB_SELECTION_FIELDS_MAP = {
    constants.LB_ALGORITHM_SOURCE_IP_PORT: ["ip_dst", "ip_src",
                                            "tp_dst", "tp_src"],
    constants.LB_ALGORITHM_SOURCE_IP: ["ip_src", "ip_dst"],
    None: ["ip_src", "ip_dst", "tp_src", "tp_dst"],
}

# HM events status
HM_EVENT_MEMBER_PORT_ONLINE = ['online']
HM_EVENT_MEMBER_PORT_OFFLINE = ['offline']

# max timeout for request
MAX_TIMEOUT_REQUEST = 5

AFFINITY_TIMEOUT = "affinity_timeout"
# This driver only supports SOURCE_IP sesssion persistency option
OVN_NATIVE_SESSION_PERSISTENCE = [constants.SESSION_PERSISTENCE_SOURCE_IP]



# Copyright 2022 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from neutron.tests import base

from ovn_octavia_provider.common import config as ovn_config
from ovn_octavia_provider.common import utils


class TestRetryDecorator(base.BaseTestCase):
    DEFAULT_RETRY_VALUE = 10

    def setUp(self):
        super().setUp()
        mock.patch.object(
            ovn_config, "get_ovn_ovsdb_retry_max_interval",
            return_value=self.DEFAULT_RETRY_VALUE).start()

    def test_default_retry_value(self):
        with mock.patch('tenacity.wait_exponential') as m_wait:
            @utils.retry()
            def decorated_method():
                pass

            decorated_method()
        m_wait.assert_called_with(max=self.DEFAULT_RETRY_VALUE)

    def test_custom_retry_value(self):
        custom_value = 3
        with mock.patch('tenacity.wait_exponential') as m_wait:
            @utils.retry(max_=custom_value)
            def decorated_method():
                pass

            decorated_method()
        m_wait.assert_called_with(max=custom_value)

    def test_positive_result(self):
        number_of_exceptions = 3
        method = mock.Mock(
            side_effect=[Exception() for i in range(number_of_exceptions)])

        @utils.retry(max_=0.001)
        def decorated_method():
            try:
                method()
            except StopIteration:
                return

        decorated_method()

        # number of exceptions + one successful call
        self.assertEqual(number_of_exceptions + 1, method.call_count)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest import mock

from neutron.tests import base
from octavia_lib.api.drivers import driver_lib
from oslo_utils import uuidutils


class TestOvnOctaviaBase(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.listener_id = uuidutils.generate_uuid()
        self.loadbalancer_id = uuidutils.generate_uuid()
        self.pool_id = uuidutils.generate_uuid()
        self.member_id = uuidutils.generate_uuid()
        self.member_subnet_id = uuidutils.generate_uuid()
        self.member_port = '1010'
        self.member_pool_id = self.pool_id
        self.member_address = '192.168.2.149'
        self.port1_id = uuidutils.generate_uuid()
        self.port2_id = uuidutils.generate_uuid()
        self.project_id = uuidutils.generate_uuid()
        self.vip_network_id = uuidutils.generate_uuid()
        self.vip_port_id = uuidutils.generate_uuid()
        self.vip_subnet_id = uuidutils.generate_uuid()
        self.healthmonitor_id = uuidutils.generate_uuid()
        ovn_nb_idl = mock.patch(
            'ovn_octavia_provider.ovsdb.impl_idl_ovn.OvnNbIdlForLb')
        self.mock_ovn_nb_idl = ovn_nb_idl.start()
        ovn_sb_idl = mock.patch(
            'ovn_octavia_provider.ovsdb.impl_idl_ovn.OvnSbIdlForLb')
        self.mock_ovn_sb_idl = ovn_sb_idl.start()
        ovsdb_nb_idl = mock.patch(
            'ovn_octavia_provider.ovsdb.impl_idl_ovn.OvsdbNbOvnIdl')
        self.mock_ovsdb_nb_idl = ovsdb_nb_idl.start()
        connection = mock.patch(
            'ovsdbapp.backend.ovs_idl.connection.Connection')
        self.mock_connection = connection.start()
        self.member_address = '192.168.2.149'
        self.vip_address = '192.148.210.109'
        self.vip_dict = {'vip_network_id': uuidutils.generate_uuid(),
                         'vip_subnet_id': uuidutils.generate_uuid()}
        self.vip_output = {'vip_network_id': self.vip_dict['vip_network_id'],
                           'vip_subnet_id': self.vip_dict['vip_subnet_id']}
        self.additional_vips = [{
            'ip_address': '2001:db8:0:1::12',
            'network_id': self.vip_dict['vip_network_id'],
            'port_id': uuidutils.generate_uuid(),
            'subnet_id': uuidutils.generate_uuid()
        }]
        mock.patch(
            'ovsdbapp.backend.ovs_idl.idlutils.get_schema_helper').start()
        mock.patch.object(
            driver_lib.DriverLibrary, '_check_for_socket_ready').start()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
import sys
from unittest import mock

from oslo_config import cfg
from oslo_log import log

from ovn_octavia_provider.cmd import octavia_ovn_db_sync_util
from ovn_octavia_provider import driver
from ovn_octavia_provider.tests.unit import base as ovn_base


class TestCMD(ovn_base.TestOvnOctaviaBase):

    def setUp(self):
        super().setUp()
        mock.patch.object(log, 'register_options').start()
        self.m_cfg = mock.patch.object(
            cfg.ConfigOpts, '__call__').start()

    @mock.patch.object(driver.OvnProviderDriver, 'do_sync')
    def test_octavia_ovn_db_sync_util(self, m_sync):
        octavia_ovn_db_sync_util.main()
        m_sync.assert_called_once_with(provider='ovn')

    @mock.patch.object(cfg.CONF, 'set_override')
    @mock.patch.object(driver.OvnProviderDriver, 'do_sync')
    def test_octavia_ovn_db_sync_util_with_debug(self, m_sync, m_cfg_or):
        return_value = ['octavia-ovn-db-sync-util',
                        '--debug']
        return_value_no_debug = ['octavia-ovn-db-sync-util']
        with mock.patch.object(sys, 'argv', return_value):
            octavia_ovn_db_sync_util.main()
        with mock.patch.object(sys, 'argv', return_value_no_debug):
            octavia_ovn_db_sync_util.main()
        m_cfg_or.assert_has_calls([mock.call('debug', True),
                                   mock.call('debug', False)])

    @mock.patch.object(octavia_ovn_db_sync_util, 'LOG')
    def test_octavia_ovn_db_sync_util_config_error(self, m_log):
        self.m_cfg.side_effect = [TypeError()]
        self.assertRaises(TypeError, octavia_ovn_db_sync_util.main)
        msg = ("Error parsing the configuration values. Please verify.")
        m_log.error.assert_called_once_with(msg)
